"""Add tcf columns

Revision ID: 66df7d9b8103
Revises: 52b2cfd1af47
Create Date: 2023-08-23 21:12:43.651877

This is a large migration, but in short, we are expanding our existing tables that let us save preferences against
privacy notices to allow us to also save preferences against TCF components.  Every TCF attribute gets its own column.
"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "66df7d9b8103"
down_revision = "52b2cfd1af47"
branch_labels = None
depends_on = None


def upgrade():
    # Add new consent settings table!
    op.create_table(
        "consentsettings",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("tcf_enabled", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_consentsettings_id"), "consentsettings", ["id"], unique=False
    )
    # Add TCF components to CurrentPrivacy Preference table: these allow you to store the last saved preference
    # for a given user across TCF versions and across time

    op.add_column(
        "currentprivacypreference", sa.Column("tcf_version", sa.String(), nullable=True)
    )
    op.add_column(
        "currentprivacypreference", sa.Column("feature", sa.Integer(), nullable=True)
    )
    op.add_column(
        "currentprivacypreference", sa.Column("purpose", sa.Integer(), nullable=True)
    )
    op.add_column(
        "currentprivacypreference",
        sa.Column("special_feature", sa.Integer(), nullable=True),
    )
    op.add_column(
        "currentprivacypreference",
        sa.Column("special_purpose", sa.Integer(), nullable=True),
    )
    op.add_column(
        "currentprivacypreference", sa.Column("vendor", sa.String(), nullable=True)
    )
    op.add_column(
        "currentprivacypreference", sa.Column("system", sa.String(), nullable=True)
    )
    # Make privacy notice id and privacy notice history id nullable, as the CurrentPrivacyPreference table
    # is being expanded to save preferences against multiple constructs, not just privacy notices
    op.alter_column(
        "currentprivacypreference",
        "privacy_notice_id",
        existing_type=sa.VARCHAR(),
        nullable=True,
    )
    op.alter_column(
        "currentprivacypreference",
        "privacy_notice_history_id",
        existing_type=sa.VARCHAR(),
        nullable=True,
    )
    # Create unique constraints between the identity for a fides device and a TCF component in
    # the currentprivacypreference table.  We want this table to store only one preference of each type per fides user device.

    op.create_unique_constraint(
        "fides_user_device_identity_feature",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "feature"],
    )
    op.create_unique_constraint(
        "fides_user_device_identity_purpose",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "purpose"],
    )
    op.create_unique_constraint(
        "fides_user_device_identity_special_feature",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "special_feature"],
    )
    op.create_unique_constraint(
        "fides_user_device_identity_special_purpose",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "special_purpose"],
    )
    op.create_unique_constraint(
        "fides_user_device_identity_system",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "system"],
    )
    op.create_unique_constraint(
        "fides_user_device_identity_vendor",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "vendor"],
    )
    # Create unique constraints between a provided identity (like an email/phone) and a TCF component in
    # the currentprivacypreference table.  We want this table to store only one preference of each type per provided identity.

    op.create_unique_constraint(
        "identity_feature",
        "currentprivacypreference",
        ["provided_identity_id", "feature"],
    )
    op.create_unique_constraint(
        "identity_purpose",
        "currentprivacypreference",
        ["provided_identity_id", "purpose"],
    )
    op.create_unique_constraint(
        "identity_special_feature",
        "currentprivacypreference",
        ["provided_identity_id", "special_feature"],
    )
    op.create_unique_constraint(
        "identity_special_purpose",
        "currentprivacypreference",
        ["provided_identity_id", "purpose"],
    )
    op.create_unique_constraint(
        "identity_system",
        "currentprivacypreference",
        ["provided_identity_id", "system"],
    )
    op.create_unique_constraint(
        "identity_vendor",
        "currentprivacypreference",
        ["provided_identity_id", "vendor"],
    )
    # Now we index these columns to make it easier to retrieve the latest preference or filter in the future
    op.create_index(
        op.f("ix_currentprivacypreference_feature"),
        "currentprivacypreference",
        ["feature"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_purpose"),
        "currentprivacypreference",
        ["purpose"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_special_feature"),
        "currentprivacypreference",
        ["special_feature"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_special_purpose"),
        "currentprivacypreference",
        ["special_purpose"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_system"),
        "currentprivacypreference",
        ["system"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_vendor"),
        "currentprivacypreference",
        ["vendor"],
        unique=False,
    )

    # Similar to the above, add columns to the existing "last served notice" table to be able to store when a
    # a TCF component was last served to a given user

    op.add_column(
        "lastservednotice", sa.Column("tcf_version", sa.String(), nullable=True)
    )
    op.add_column("lastservednotice", sa.Column("feature", sa.Integer(), nullable=True))
    op.add_column("lastservednotice", sa.Column("purpose", sa.Integer(), nullable=True))
    op.add_column(
        "lastservednotice", sa.Column("special_feature", sa.Integer(), nullable=True)
    )
    op.add_column(
        "lastservednotice", sa.Column("special_purpose", sa.Integer(), nullable=True)
    )
    op.add_column("lastservednotice", sa.Column("vendor", sa.String(), nullable=True))
    op.add_column("lastservednotice", sa.Column("system", sa.String(), nullable=True))

    # Relax a criteria that we had to have a privacy notice id and privacy notice history id to store that consent
    # was served. The Last Served Notice table is now shared with TCF components.
    op.alter_column(
        "lastservednotice",
        "privacy_notice_id",
        existing_type=sa.VARCHAR(),
        nullable=True,
    )
    op.alter_column(
        "lastservednotice",
        "privacy_notice_history_id",
        existing_type=sa.VARCHAR(),
        nullable=True,
    )

    # Add unique constraints between a fides user device id and the tcf component.  For each device and particular TCF
    # component, we want exactly one record stored in this table.  This records the last time consent was served!
    op.create_unique_constraint(
        "last_served_fides_user_device_identity_feature",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "feature"],
    )
    op.create_unique_constraint(
        "last_served_fides_user_device_identity_purpose",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "purpose"],
    )
    op.create_unique_constraint(
        "last_served_fides_user_device_identity_special_feature",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "special_feature"],
    )
    op.create_unique_constraint(
        "last_served_fides_user_device_identity_special_purpose",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "special_purpose"],
    )
    op.create_unique_constraint(
        "last_served_fides_user_device_identity_system",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "system"],
    )
    op.create_unique_constraint(
        "last_served_fides_user_device_identity_vendor",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "vendor"],
    )
    # Add unique constraints between a provided identity like an email/phone and the tcf component.  For each provided identity
    # and particular TCF component, we want exactly one record stored in this table.  This records the last time consent was served!
    op.create_unique_constraint(
        "last_served_identity_feature",
        "lastservednotice",
        ["provided_identity_id", "feature"],
    )
    op.create_unique_constraint(
        "last_served_identity_purpose",
        "lastservednotice",
        ["provided_identity_id", "purpose"],
    )
    op.create_unique_constraint(
        "last_served_identity_special_feature",
        "lastservednotice",
        ["provided_identity_id", "special_feature"],
    )
    op.create_unique_constraint(
        "last_served_identity_special_purpose",
        "lastservednotice",
        ["provided_identity_id", "special_purpose"],
    )
    op.create_unique_constraint(
        "last_served_identity_system",
        "lastservednotice",
        ["provided_identity_id", "system"],
    )
    op.create_unique_constraint(
        "last_served_identity_vendor",
        "lastservednotice",
        ["provided_identity_id", "vendor"],
    )
    # Lastly, let's index these new TCF columns on the LastServedNotice table
    op.create_index(
        op.f("ix_lastservednotice_feature"),
        "lastservednotice",
        ["feature"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_purpose"),
        "lastservednotice",
        ["purpose"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_special_feature"),
        "lastservednotice",
        ["special_feature"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_special_purpose"),
        "lastservednotice",
        ["special_purpose"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_system"), "lastservednotice", ["system"], unique=False
    )
    op.create_index(
        op.f("ix_lastservednotice_vendor"), "lastservednotice", ["vendor"], unique=False
    )
    # Add new TCF columns to the PrivacyPreferenceHistory table to let us save each time consent is served
    # against a TCF component.
    op.add_column(
        "privacypreferencehistory", sa.Column("feature", sa.Integer(), nullable=True)
    )
    op.add_column(
        "privacypreferencehistory", sa.Column("purpose", sa.Integer(), nullable=True)
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("special_feature", sa.Integer(), nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("special_purpose", sa.Integer(), nullable=True),
    )
    op.add_column(
        "privacypreferencehistory", sa.Column("vendor", sa.String(), nullable=True)
    )
    op.add_column(
        "privacypreferencehistory", sa.Column("system", sa.String(), nullable=True)
    )
    op.add_column(
        "privacypreferencehistory", sa.Column("tcf_version", sa.String(), nullable=True)
    )
    # Relax criteria so we can save privacy preference history against more than just privacy notices.
    op.alter_column(
        "privacypreferencehistory",
        "privacy_notice_history_id",
        existing_type=sa.VARCHAR(),
        nullable=True,
    )
    # Index the new TCF columns to make them easier to search and filter on
    op.create_index(
        op.f("ix_privacypreferencehistory_feature"),
        "privacypreferencehistory",
        ["feature"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_purpose"),
        "privacypreferencehistory",
        ["purpose"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_special_feature"),
        "privacypreferencehistory",
        ["special_feature"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_special_purpose"),
        "privacypreferencehistory",
        ["special_purpose"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_system"),
        "privacypreferencehistory",
        ["system"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_vendor"),
        "privacypreferencehistory",
        ["vendor"],
        unique=False,
    )
    # Add new TCF columns to the ServedNoticeHistory table.  This lets us record every time TCF components
    # are served to an end user.s
    op.add_column(
        "servednoticehistory", sa.Column("feature", sa.Integer(), nullable=True)
    )
    op.add_column(
        "servednoticehistory", sa.Column("purpose", sa.Integer(), nullable=True)
    )
    op.add_column(
        "servednoticehistory", sa.Column("special_feature", sa.Integer(), nullable=True)
    )
    op.add_column(
        "servednoticehistory", sa.Column("special_purpose", sa.Integer(), nullable=True)
    )
    op.add_column(
        "servednoticehistory", sa.Column("vendor", sa.String(), nullable=True)
    )
    op.add_column(
        "servednoticehistory", sa.Column("system", sa.String(), nullable=True)
    )
    op.add_column(
        "servednoticehistory", sa.Column("tcf_version", sa.String(), nullable=True)
    )
    # Relax constraints so we can persist that TCF components were served. Not just privacy notices.
    op.alter_column(
        "servednoticehistory",
        "privacy_notice_history_id",
        existing_type=sa.VARCHAR(),
        nullable=True,
    )
    # Index the new TCF columns in ServedNoticeHistory
    op.create_index(
        op.f("ix_servednoticehistory_feature"),
        "servednoticehistory",
        ["feature"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_purpose"),
        "servednoticehistory",
        ["purpose"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_special_feature"),
        "servednoticehistory",
        ["special_feature"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_special_purpose"),
        "servednoticehistory",
        ["special_purpose"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_system"),
        "servednoticehistory",
        ["system"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_vendor"),
        "servednoticehistory",
        ["vendor"],
        unique=False,
    )


def downgrade():
    op.drop_index(
        op.f("ix_servednoticehistory_vendor"), table_name="servednoticehistory"
    )
    op.drop_index(
        op.f("ix_servednoticehistory_system"), table_name="servednoticehistory"
    )
    op.drop_index(
        op.f("ix_servednoticehistory_special_purpose"), table_name="servednoticehistory"
    )
    op.drop_index(
        op.f("ix_servednoticehistory_special_feature"), table_name="servednoticehistory"
    )
    op.drop_index(
        op.f("ix_servednoticehistory_purpose"), table_name="servednoticehistory"
    )
    op.drop_index(
        op.f("ix_servednoticehistory_feature"), table_name="servednoticehistory"
    )
    op.drop_column("servednoticehistory", "tcf_version")
    op.drop_column("servednoticehistory", "system")
    op.drop_column("servednoticehistory", "vendor")
    op.drop_column("servednoticehistory", "special_purpose")
    op.drop_column("servednoticehistory", "special_feature")
    op.drop_column("servednoticehistory", "purpose")
    op.drop_column("servednoticehistory", "feature")
    op.drop_index(
        op.f("ix_privacypreferencehistory_vendor"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_system"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_special_purpose"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_special_feature"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_purpose"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_feature"),
        table_name="privacypreferencehistory",
    )
    op.drop_column("privacypreferencehistory", "tcf_version")
    op.drop_column("privacypreferencehistory", "system")
    op.drop_column("privacypreferencehistory", "vendor")
    op.drop_column("privacypreferencehistory", "special_purpose")
    op.drop_column("privacypreferencehistory", "special_feature")
    op.drop_column("privacypreferencehistory", "purpose")
    op.drop_column("privacypreferencehistory", "feature")

    op.drop_constraint(
        "last_served_identity_vendor", "lastservednotice", type_="unique"
    )
    op.drop_constraint(
        "last_served_identity_system", "lastservednotice", type_="unique"
    )
    op.drop_constraint(
        "last_served_identity_special_purpose", "lastservednotice", type_="unique"
    )
    op.drop_constraint(
        "last_served_identity_special_feature", "lastservednotice", type_="unique"
    )
    op.drop_constraint(
        "last_served_identity_purpose", "lastservednotice", type_="unique"
    )
    op.drop_constraint(
        "last_served_identity_feature", "lastservednotice", type_="unique"
    )
    op.drop_constraint(
        "last_served_fides_user_device_identity_vendor",
        "lastservednotice",
        type_="unique",
    )
    op.drop_constraint(
        "last_served_fides_user_device_identity_system",
        "lastservednotice",
        type_="unique",
    )
    op.drop_constraint(
        "last_served_fides_user_device_identity_special_purpose",
        "lastservednotice",
        type_="unique",
    )
    op.drop_constraint(
        "last_served_fides_user_device_identity_special_feature",
        "lastservednotice",
        type_="unique",
    )
    op.drop_constraint(
        "last_served_fides_user_device_identity_purpose",
        "lastservednotice",
        type_="unique",
    )
    op.drop_constraint(
        "last_served_fides_user_device_identity_feature",
        "lastservednotice",
        type_="unique",
    )
    op.drop_index(op.f("ix_lastservednotice_vendor"), table_name="lastservednotice")
    op.drop_index(op.f("ix_lastservednotice_system"), table_name="lastservednotice")
    op.drop_index(
        op.f("ix_lastservednotice_special_purpose"), table_name="lastservednotice"
    )
    op.drop_index(
        op.f("ix_lastservednotice_special_feature"), table_name="lastservednotice"
    )
    op.drop_index(op.f("ix_lastservednotice_purpose"), table_name="lastservednotice")
    op.drop_index(op.f("ix_lastservednotice_feature"), table_name="lastservednotice")
    op.drop_column("lastservednotice", "system")
    op.drop_column("lastservednotice", "vendor")
    op.drop_column("lastservednotice", "special_purpose")
    op.drop_column("lastservednotice", "special_feature")
    op.drop_column("lastservednotice", "purpose")
    op.drop_column("lastservednotice", "feature")
    op.drop_column("lastservednotice", "tcf_version")
    op.drop_index(
        op.f("ix_currentprivacypreference_vendor"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_system"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_special_purpose"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_special_feature"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_purpose"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_feature"),
        table_name="currentprivacypreference",
    )
    op.drop_constraint("identity_vendor", "currentprivacypreference", type_="unique")
    op.drop_constraint("identity_system", "currentprivacypreference", type_="unique")
    op.drop_constraint(
        "identity_special_purpose", "currentprivacypreference", type_="unique"
    )
    op.drop_constraint(
        "identity_special_feature", "currentprivacypreference", type_="unique"
    )
    op.drop_constraint("identity_purpose", "currentprivacypreference", type_="unique")
    op.drop_constraint("identity_feature", "currentprivacypreference", type_="unique")
    op.drop_constraint(
        "fides_user_device_identity_vendor", "currentprivacypreference", type_="unique"
    )
    op.drop_constraint(
        "fides_user_device_identity_system", "currentprivacypreference", type_="unique"
    )
    op.drop_constraint(
        "fides_user_device_identity_special_purpose",
        "currentprivacypreference",
        type_="unique",
    )
    op.drop_constraint(
        "fides_user_device_identity_special_feature",
        "currentprivacypreference",
        type_="unique",
    )
    op.drop_constraint(
        "fides_user_device_identity_purpose", "currentprivacypreference", type_="unique"
    )
    op.drop_constraint(
        "fides_user_device_identity_feature", "currentprivacypreference", type_="unique"
    )
    op.drop_column("currentprivacypreference", "system")
    op.drop_column("currentprivacypreference", "vendor")
    op.drop_column("currentprivacypreference", "special_purpose")
    op.drop_column("currentprivacypreference", "special_feature")
    op.drop_column("currentprivacypreference", "purpose")
    op.drop_column("currentprivacypreference", "feature")
    op.drop_column("currentprivacypreference", "tcf_version")
    op.drop_index(op.f("ix_consentsettings_id"), table_name="consentsettings")
    op.drop_table("consentsettings")



"""add user_id to manual_task_log

Revision ID: 641f6bcd424f
Revises: 41a634d8c669
Create Date: 2025-07-02 03:19:40.171201

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "641f6bcd424f"
down_revision = "41a634d8c669"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("manual_task_log", sa.Column("user_id", sa.String(), nullable=True))
    op.create_index(
        op.f("ix_manual_task_log_user_id"), "manual_task_log", ["user_id"], unique=False
    )


def downgrade():
    op.drop_index(op.f("ix_manual_task_log_user_id"), table_name="manual_task_log")
    op.drop_column("manual_task_log", "user_id")



"""add_manual_task_instance_table

Revision ID: 6a76a1fa4f3f
Revises: ba414a58ba90
Create Date: 2025-06-10 22:45:27.591492

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import JSONB

# revision identifiers, used by Alembic.
revision = "6a76a1fa4f3f"
down_revision = "ba414a58ba90"
branch_labels = None
depends_on = None


def upgrade():
    # Add execution_timing column to manual_task_config
    op.add_column(
        "manual_task_config",
        sa.Column(
            "execution_timing",
            sa.String(),
            nullable=False,
            server_default="pre_execution",
        ),
    )

    # Remove due_date column from manual_task
    op.drop_column("manual_task", "due_date")

    # Create manual_task_instance table
    op.create_table(
        "manual_task_instance",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column("task_id", sa.String(), nullable=False),
        sa.Column("config_id", sa.String(), nullable=False),
        sa.Column("entity_id", sa.String(), nullable=False),
        sa.Column("entity_type", sa.String(), nullable=False),
        sa.Column("status", sa.String(), nullable=False, server_default="pending"),
        sa.Column("due_date", sa.DateTime(timezone=True), nullable=True),
        sa.Column("completed_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("completed_by_id", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["config_id"], ["manual_task_config.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(["task_id"], ["manual_task.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )

    # Create indexes for manual_task_instance
    op.create_index(
        "ix_manual_task_instance_task_id", "manual_task_instance", ["task_id"]
    )
    op.create_index(
        "ix_manual_task_instance_config_id", "manual_task_instance", ["config_id"]
    )
    op.create_index(
        "ix_manual_task_instance_entity_id", "manual_task_instance", ["entity_id"]
    )
    op.create_index(
        "ix_manual_task_instance_entity_type", "manual_task_instance", ["entity_type"]
    )
    op.create_index(
        "ix_manual_task_instance_status", "manual_task_instance", ["status"]
    )
    op.create_index(
        "ix_manual_task_instance_completed_at", "manual_task_instance", ["completed_at"]
    )
    # Composite index for common query pattern
    op.create_index(
        "ix_manual_task_instance_entity",
        "manual_task_instance",
        ["entity_type", "entity_id"],
    )

    # Create manual_task_submission table
    op.create_table(
        "manual_task_submission",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column("task_id", sa.String(), nullable=False),
        sa.Column("config_id", sa.String(), nullable=False),
        sa.Column("field_id", sa.String(), nullable=False),
        sa.Column("instance_id", sa.String(), nullable=False),
        sa.Column("submitted_by", sa.String(), nullable=True),
        sa.Column("submitted_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("data", JSONB, nullable=False),
        sa.ForeignKeyConstraint(
            ["config_id"], ["manual_task_config.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["field_id"], ["manual_task_config_field.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(
            ["instance_id"], ["manual_task_instance.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(["task_id"], ["manual_task.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(
            ["submitted_by"], ["fidesuser.id"], ondelete="SET NULL"
        ),
        sa.PrimaryKeyConstraint("id"),
    )

    # Create indexes for manual_task_submission
    op.create_index(
        "ix_manual_task_submission_task_id", "manual_task_submission", ["task_id"]
    )
    op.create_index(
        "ix_manual_task_submission_config_id", "manual_task_submission", ["config_id"]
    )
    op.create_index(
        "ix_manual_task_submission_field_id", "manual_task_submission", ["field_id"]
    )
    op.create_index(
        "ix_manual_task_submission_instance_id",
        "manual_task_submission",
        ["instance_id"],
    )
    op.create_index(
        "ix_manual_task_submission_submitted_by",
        "manual_task_submission",
        ["submitted_by"],
    )
    op.create_index(
        "ix_manual_task_submission_submitted_at",
        "manual_task_submission",
        ["submitted_at"],
    )
    # Composite index for common query pattern
    op.create_index(
        "ix_manual_task_submission_instance_field",
        "manual_task_submission",
        ["instance_id", "field_id"],
    )

    # Add foreign key constraint to manual_task_log.instance_id
    op.create_foreign_key(
        "fk_manual_task_log_instance_id",
        "manual_task_log",
        "manual_task_instance",
        ["instance_id"],
        ["id"],
        ondelete="CASCADE",
    )

    # Update foreign key constraint for manual_task_log.config_id to use CASCADE
    op.drop_constraint(
        "fk_manual_task_log_config_id", "manual_task_log", type_="foreignkey"
    )
    op.create_foreign_key(
        "fk_manual_task_log_config_id",
        "manual_task_log",
        "manual_task_config",
        ["config_id"],
        ["id"],
        ondelete="CASCADE",
    )


def downgrade():
    # Drop execution_timing column from manual_task_config
    op.drop_column("manual_task_config", "execution_timing")

    # Add due_date column to manual_task
    op.add_column(
        "manual_task",
        sa.Column("due_date", sa.DateTime(timezone=True), nullable=True),
    )

    # Drop foreign key constraint from manual_task_log.instance_id
    op.drop_constraint(
        "fk_manual_task_log_instance_id", "manual_task_log", type_="foreignkey"
    )

    # Revert foreign key constraint for manual_task_log.config_id back to SET NULL
    op.drop_constraint(
        "fk_manual_task_log_config_id", "manual_task_log", type_="foreignkey"
    )
    op.create_foreign_key(
        "fk_manual_task_log_config_id",
        "manual_task_log",
        "manual_task_config",
        ["config_id"],
        ["id"],
        ondelete="SET NULL",
    )

    # Drop indexes first
    op.drop_index(
        "ix_manual_task_submission_instance_field", table_name="manual_task_submission"
    )
    op.drop_index(
        "ix_manual_task_submission_submitted_at", table_name="manual_task_submission"
    )
    op.drop_index(
        "ix_manual_task_submission_submitted_by", table_name="manual_task_submission"
    )
    op.drop_index(
        "ix_manual_task_submission_instance_id", table_name="manual_task_submission"
    )
    op.drop_index(
        "ix_manual_task_submission_field_id", table_name="manual_task_submission"
    )
    op.drop_index(
        "ix_manual_task_submission_config_id", table_name="manual_task_submission"
    )
    op.drop_index(
        "ix_manual_task_submission_task_id", table_name="manual_task_submission"
    )

    op.drop_index("ix_manual_task_instance_entity", table_name="manual_task_instance")
    op.drop_index(
        "ix_manual_task_instance_completed_at", table_name="manual_task_instance"
    )
    op.drop_index("ix_manual_task_instance_status", table_name="manual_task_instance")
    op.drop_index(
        "ix_manual_task_instance_entity_type", table_name="manual_task_instance"
    )
    op.drop_index(
        "ix_manual_task_instance_entity_id", table_name="manual_task_instance"
    )
    op.drop_index(
        "ix_manual_task_instance_config_id", table_name="manual_task_instance"
    )
    op.drop_index("ix_manual_task_instance_task_id", table_name="manual_task_instance")

    # Drop tables in reverse order to handle foreign key constraints
    op.drop_table("manual_task_submission")
    op.drop_table("manual_task_instance")



"""Add banner_description to experience config

Revision ID: f396c1f84b0f
Revises: 4ced99dabebb
Create Date: 2024-01-02 01:35:41.813279

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f396c1f84b0f"
down_revision = "4ced99dabebb"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("banner_description", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("banner_description", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("banner_title", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("banner_title", sa.String(), nullable=True),
    )


def downgrade():
    op.drop_column("privacyexperienceconfighistory", "banner_title")
    op.drop_column("privacyexperienceconfig", "banner_title")
    op.drop_column("privacyexperienceconfighistory", "banner_description")
    op.drop_column("privacyexperienceconfig", "banner_description")



"""classification management

Revision ID: 3d9c476d7cea
Revises: 6bd93cb0603d
Create Date: 2022-09-14 02:21:37.828382

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3d9c476d7cea"
down_revision = "6bd93cb0603d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "cls_classification_detail",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column("instance_id", sa.String(length=255), nullable=True),
        sa.Column("status", sa.Text(), nullable=True),
        sa.Column("dataset", sa.Text(), nullable=True),
        sa.Column("collection", sa.Text(), nullable=True),
        sa.Column("field", sa.Text(), nullable=True),
        sa.Column("labels", sa.JSON(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_cls_classification_detail_id"),
        "cls_classification_detail",
        ["id"],
        unique=False,
    )
    op.create_table(
        "cls_classification_instance",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column("status", sa.Text(), nullable=True),
        sa.Column("organization_key", sa.Text(), nullable=True),
        sa.Column("dataset_key", sa.Text(), nullable=True),
        sa.Column("dataset_name", sa.Text(), nullable=True),
        sa.Column("target", sa.Text(), nullable=True),
        sa.Column("type", sa.Text(), nullable=True),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_cls_classification_instance_id"),
        "cls_classification_instance",
        ["id"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_cls_classification_instance_id"),
        table_name="cls_classification_instance",
    )
    op.drop_table("cls_classification_instance")
    op.drop_index(
        op.f("ix_cls_classification_detail_id"), table_name="cls_classification_detail"
    )
    op.drop_table("cls_classification_detail")
    # ### end Alembic commands ###



from typing import List

from fastapi import HTTPException, Security, status
from loguru import logger
from sqlalchemy_bigquery import BigQueryDialect

from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.partitioning.time_based_partitioning import (
    TimeBasedPartitioning,
    combine_partitions,
)
from fides.api.util.api_router import APIRouter
from fides.common.api.scope_registry import CTL_DATASET_READ
from fides.common.api.v1.urn_registry import V1_URL_PREFIX

router = APIRouter(tags=["Partitions"], prefix=V1_URL_PREFIX)


@router.post(
    "/dataset/partitions/verify",
    dependencies=[Security(verify_oauth_client, scopes=[CTL_DATASET_READ])],
    response_model=List[str],
    status_code=status.HTTP_200_OK,
    name="Verify dataset partitioning schemas",
    summary="Verify dataset partitioning schemas",
    description="Validate the provided partitioning specs and return the rendered SQL WHERE clauses in the BigQuery dialect.",
    openapi_extra={
        "requestBody": {
            "content": {
                "application/json": {
                    "schema": {
                        "type": "array",
                        "items": {"$ref": "#/components/schemas/TimeBasedPartitioning"},
                    },
                    "examples": {
                        "recent_data_weekly": {
                            "summary": "Recent data partitioned weekly",
                            "description": "Partition the last 30 days of data into weekly chunks",
                            "value": [
                                {
                                    "field": "created_at",
                                    "start": "NOW() - 30 DAYS",
                                    "end": "NOW()",
                                    "interval": "7 DAYS",
                                }
                            ],
                        },
                        "historical_data_monthly": {
                            "summary": "Historical data partitioned monthly",
                            "description": "Partition a specific year of data into monthly chunks",
                            "value": [
                                {
                                    "field": "order_date",
                                    "start": "2024-01-01",
                                    "end": "2024-12-31",
                                    "interval": "1 MONTH",
                                }
                            ],
                        },
                        "multiple_partitions": {
                            "summary": "Multiple time ranges",
                            "description": "Combine historical monthly data with recent daily data",
                            "value": [
                                {
                                    "field": "event_timestamp",
                                    "start": "2023-01-01",
                                    "end": "2023-12-31",
                                    "interval": "1 MONTH",
                                },
                                {
                                    "field": "event_timestamp",
                                    "start": "NOW() - 14 DAYS",
                                    "end": "NOW()",
                                    "interval": "1 DAY",
                                },
                            ],
                        },
                        "open_ended_partition": {
                            "summary": "Open-ended partition",
                            "description": "Partition all data from a specific date onwards",
                            "value": [
                                {"field": "user_created_at", "start": "2024-01-01"}
                            ],
                        },
                    },
                }
            }
        }
    },
)
async def verify_dataset_partitioning(
    partitions: List[TimeBasedPartitioning],
) -> List[str]:
    """Validate the provided partitioning specs and return the rendered SQL WHERE clauses in the BigQuery dialect."""
    try:
        expressions = combine_partitions(partitions)
        dialect = BigQueryDialect()
        compiled_clauses: List[str] = [
            str(expr.compile(dialect=dialect, compile_kwargs={"literal_binds": True}))
            for expr in expressions
        ]
        return compiled_clauses
    except Exception as exc:
        logger.error("Partitioning verification failed", exc_info=True)
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=str(exc),
        )



from typing import Optional

from google.cloud.storage import Blob, Client  # type: ignore
from google.oauth2 import service_account
from loguru import logger

from fides.api.common_exceptions import StorageUploadError
from fides.api.schemas.storage.storage import GCSAuthMethod


def get_gcs_client(
    auth_method: str,
    storage_secrets: Optional[dict],
) -> Client:
    """
    Abstraction to retrieve a GCS client using secrets.
    """
    if auth_method == GCSAuthMethod.ADC.value:
        storage_client = Client()

    elif auth_method == GCSAuthMethod.SERVICE_ACCOUNT_KEYS.value:
        if not storage_secrets:
            err_msg = "Storage secrets not found for Google Cloud Storage."
            logger.warning(err_msg)
            raise StorageUploadError(err_msg)

        credentials = service_account.Credentials.from_service_account_info(
            dict(storage_secrets)
        )
        storage_client = Client(credentials=credentials)

    else:
        logger.error("Google Cloud Storage auth method not supported: {}", auth_method)
        raise ValueError(
            f"Google Cloud Storage auth method not supported: {auth_method}"
        )

    return storage_client


def get_gcs_blob(
    auth_method: str, storage_secrets: Optional[dict], bucket_name: str, file_key: str
) -> Blob:
    try:
        storage_client = get_gcs_client(auth_method, storage_secrets)
        bucket = storage_client.bucket(bucket_name)
        return bucket.blob(file_key)
    except Exception as e:
        logger.error(f"Error getting GCS blob: {str(e)}")
        raise e



"""create manual task config tables

Revision ID: ba414a58ba90
Revises: 29e56fa1fdb3
Create Date: 2025-06-09 19:56:18.254209

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import JSONB

# revision identifiers, used by Alembic.
revision = "ba414a58ba90"
down_revision = "29e56fa1fdb3"
branch_labels = None
depends_on = None


def upgrade():
    # Create manual_task_config table
    op.create_table(
        "manual_task_config",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            nullable=False,
            server_default=sa.text("now()"),
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            nullable=False,
            server_default=sa.text("now()"),
        ),
        sa.Column("task_id", sa.String(), nullable=False),
        sa.Column("config_type", sa.String(), nullable=False),
        sa.Column("version", sa.Integer(), nullable=False, server_default="1"),
        sa.Column("is_current", sa.Boolean(), nullable=False, server_default="true"),
        sa.ForeignKeyConstraint(["task_id"], ["manual_task.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )

    # Create manual_task_config_field table
    op.create_table(
        "manual_task_config_field",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            nullable=False,
            server_default=sa.text("now()"),
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            nullable=False,
            server_default=sa.text("now()"),
        ),
        sa.Column("task_id", sa.String(), nullable=False),
        sa.Column("config_id", sa.String(), nullable=False),
        sa.Column("field_key", sa.String(), nullable=False),
        sa.Column("field_type", sa.String(), nullable=False),
        sa.Column("field_metadata", JSONB, nullable=False, server_default="{}"),
        sa.ForeignKeyConstraint(["task_id"], ["manual_task.id"], ondelete="CASCADE"),
        sa.ForeignKeyConstraint(
            ["config_id"], ["manual_task_config.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "config_id", "field_key", name="unique_field_key_per_config"
        ),
    )

    # Add foreign key constraint to manual_task_log table
    op.create_foreign_key(
        "fk_manual_task_log_config_id",
        "manual_task_log",
        "manual_task_config",
        ["config_id"],
        ["id"],
        ondelete="SET NULL",
    )

    # Create indexes
    op.create_index("ix_manual_task_config_task_id", "manual_task_config", ["task_id"])
    op.create_index(
        "ix_manual_task_config_config_type", "manual_task_config", ["config_type"]
    )
    op.create_index(
        "ix_manual_task_config_field_task_id", "manual_task_config_field", ["task_id"]
    )
    op.create_index(
        "ix_manual_task_config_field_config_id",
        "manual_task_config_field",
        ["config_id"],
    )
    op.create_index(
        "ix_manual_task_config_field_field_key",
        "manual_task_config_field",
        ["field_key"],
    )


def downgrade():
    # Drop indexes
    op.drop_index("ix_manual_task_config_field_field_key")
    op.drop_index("ix_manual_task_config_field_config_id")
    op.drop_index("ix_manual_task_config_field_task_id")
    op.drop_index("ix_manual_task_config_config_type")
    op.drop_index("ix_manual_task_config_task_id")

    # Drop foreign key constraint from manual_task_log
    op.drop_constraint(
        "fk_manual_task_log_config_id", "manual_task_log", type_="foreignkey"
    )

    # Drop tables
    op.drop_table("manual_task_config_field")
    op.drop_table("manual_task_config")



from fideslang.models import Dataset as FideslangDataset

from fides.api.common_exceptions import SaaSConfigNotFoundException, ValidationError
from fides.api.models.connectionconfig import ConnectionConfig, ConnectionType
from fides.api.models.datasetconfig import to_graph_field
from fides.service.dataset.dataset_validator import (
    DatasetValidationContext,
    DatasetValidationStep,
)


def _validate_saas_dataset(
    connection_config: ConnectionConfig, dataset: FideslangDataset
) -> None:
    if connection_config.saas_config is None:
        raise SaaSConfigNotFoundException(
            f"Connection config '{connection_config.key}' must have a "
            "SaaS config before validating or adding a dataset"
        )

    fides_key = connection_config.saas_config["fides_key"]
    if fides_key != dataset.fides_key:
        raise ValidationError(
            f"The fides_key '{dataset.fides_key}' of the dataset "
            f"does not match the fides_key '{fides_key}' "
            "of the connection config"
        )
    for collection in dataset.collections:
        for field in collection.fields:
            graph_field = to_graph_field(field)
            if graph_field.references or graph_field.identity:
                raise ValidationError(
                    "A dataset for a ConnectionConfig type of 'saas' is not "
                    "allowed to have references or identities. Please add "
                    "them to the SaaS config."
                )


class SaaSValidationStep(DatasetValidationStep):
    """Validates SaaS-specific requirements"""

    def validate(self, context: DatasetValidationContext) -> None:
        if (
            context.connection_config
            and context.connection_config.connection_type == ConnectionType.saas
        ):
            _validate_saas_dataset(context.connection_config, context.dataset)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set

from fides.api.common_exceptions import TraversalError, UnreachableNodesError
from fides.api.graph.graph import DatasetGraph
from fides.api.models.policy import Policy
from fides.api.util.logger import suppress_logging

if TYPE_CHECKING:
    from fides.api.graph.traversal import TraversalNode


class NodeFilter(ABC):
    """Abstract base class for node filters that determine if nodes should be excluded from traversal"""

    def __init__(self) -> None:
        self.skipped_nodes: Dict[str, str] = {}

    @abstractmethod
    def exclude_node(self, node: "TraversalNode") -> bool:
        """Returns True if the node should be excluded from traversal requirements"""


class CustomRequestFieldFilter(NodeFilter):
    """
    Remove nodes that have custom request fields, since we don't care if these are reachable or not.
    They will be used independently by the Dynamic Email Erasure Connector.

    TODO: ideally we'll update the Traversal code to include these "custom request field datasets"
    as part of the main graph. This is a targeted workaround for now.
    """

    def exclude_node(self, node: "TraversalNode") -> bool:
        return bool(node.node.collection.custom_request_fields())


class PolicyDataCategoryFilter(NodeFilter):
    """
    If a policy is provided, we will allow collections without relevant
    data categories to remain unreachable without raising any exceptions.
    """

    def __init__(self, policy: Optional[Policy]):
        super().__init__()
        self.policy = policy

    def exclude_node(self, node: "TraversalNode") -> bool:
        if not self.policy:
            return False
        return not self.policy.applies_to(node)


class OptionalIdentityFilter(NodeFilter):

    def __init__(self, graph: DatasetGraph, identities: Dict[str, Any]):
        """
        Filter that excludes nodes that are reachable by an optional identity.
        This prevents traversal errors for nodes that would be reachable if the required identities were provided.
        """

        super().__init__()
        self.graph = graph
        self.identities = identities
        self.skipped_nodes: Dict[str, str] = {}

        # Pre-compute which nodes are reachable with each identity
        self.reachable_by_identity: Dict[str, Set[str]] = {}

        # Get all collection addresses as strings
        all_collection_addresses = {str(addr) for addr in graph.nodes.keys()}

        # Test reachability with each possible identity seed
        for identity_key in graph.identity_keys.values():
            self._compute_reachable_nodes(identity_key, all_collection_addresses)

    def exclude_node(self, node: "TraversalNode") -> bool:
        """
        Exclude nodes that are reachable by any identity, provided or optional.
        Nodes (collections) not reachable by any identities should not be excluded and still be considered an error.
        """
        node_address = str(node.address)

        # Check if node is reachable by any provided identity
        for identity_key in self.identities:
            if node_address in self.reachable_by_identity.get(identity_key, set()):
                return True

        # Find all optional identities that can reach this node
        reaching_identities = []
        for identity_key, reachable_nodes in self.reachable_by_identity.items():
            if identity_key not in self.identities and node_address in reachable_nodes:
                reaching_identities.append(identity_key)

        # If any optional identities can reach this node, record them and exclude
        if reaching_identities:
            self.skipped_nodes[node_address] = self._format_skip_message(
                node.address.value, reaching_identities
            )
            return True

        # Node isn't reachable by any identity, don't exclude it (will be an error)
        return False

    def _compute_reachable_nodes(
        self, identity_key: str, all_addresses: Set[str]
    ) -> None:
        """Helper method to compute which nodes are reachable with a given identity key"""

        from fides.api.graph.traversal import BaseTraversal

        try:
            # Create a traversal object with just this identity
            with suppress_logging():
                # Suppress the logs since we don't want to flood the logs
                # with traversal info for each identity we want to evaluate
                BaseTraversal(self.graph, {identity_key: "dummy_value"})

            # If successful, all nodes are reachable
            self.reachable_by_identity[identity_key] = all_addresses
        except UnreachableNodesError as exc:
            # Get reachable nodes by removing unreachable ones from all nodes
            unreachable = set(exc.errors)
            self.reachable_by_identity[identity_key] = all_addresses - unreachable
        except TraversalError:
            # If traversal fails for other reasons, no nodes are reachable
            self.reachable_by_identity[identity_key] = set()

    def _format_skip_message(
        self, node_address: str, reaching_identities: List[str]
    ) -> str:
        """
        Format the skip message with proper grammar based on identity counts.
        """

        # Format lists of quoted identities
        reachable_ids_str = ", ".join(f'"{id}"' for id in reaching_identities)
        provided_ids = list(self.identities.keys())
        provided_ids_str = ", ".join(f'"{id}"' for id in provided_ids)

        # Determine singular/plural forms based on counts
        reachable_term = "identity" if len(reaching_identities) == 1 else "identities"
        provided_term = "identity" if len(provided_ids) == 1 else "identities"
        was_were = "was" if len(provided_ids) == 1 else "were"

        return (
            f'Skipping the "{node_address}" collection, it is reachable by the {reachable_ids_str} '
            f"{reachable_term} but only the {provided_ids_str} {provided_term} {was_were} provided"
        )



from __future__ import annotations

from enum import Enum
from typing import Any, Dict, Optional, Type

from sqlalchemy import (
    ARRAY,
    Column,
    ForeignKey,
    Index,
    String,
    func,
    insert,
    select,
    update,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import relationship

from fides.api.db.base_class import Base
from fides.api.db.util import EnumColumn
from fides.api.models.sql_models import System  # type: ignore[attr-defined]


class ConsentStatus(str, Enum):
    """
    Consent status of the asset
    """

    with_consent = "with_consent"
    without_consent = "without_consent"
    exempt = "exempt"
    unknown = "unknown"


class Asset(Base):
    """
    Web assets associated with a system
    """

    # Common attributes
    name = Column(String, index=True, nullable=False)
    asset_type = Column(String, index=True, nullable=False)
    domain = Column(String, index=True)
    parent = Column(ARRAY(String), server_default="{}", nullable=False)
    parent_domain = Column(String)
    locations = Column(ARRAY(String), server_default="{}", nullable=False)
    consent_status = Column(
        EnumColumn(ConsentStatus, native_enum=True),
        default=ConsentStatus.unknown,
        server_default="unknown",
        nullable=False,
    )
    data_uses = Column(ARRAY(String), server_default="{}", nullable=False)
    description = Column(String, nullable=True)
    page = Column(ARRAY(String), server_default="{}", nullable=False)

    # generic object to store additional attributes, specific to asset type
    meta = Column(
        MutableDict.as_mutable(JSONB),
        nullable=False,
        server_default="{}",
        default=dict,
    )

    # Browser request-specific attributes
    base_url = Column(String)

    system_id = Column(
        String, ForeignKey(System.id_field_path, ondelete="CASCADE"), index=True
    )  # If system is deleted, remove the associated assets.

    system = relationship(
        System,
        back_populates="assets",
        cascade="all,delete",
        uselist=False,
        lazy="selectin",
    )

    # we need to use an md5 of the base_url to avoid constraint/index length issues
    # and we need to use a unique index, rather than constraint, since postgresql constraints
    # do not support expressions, only direct column references
    __table_args__ = (
        Index(
            "ix_asset_name_asset_type_domain_base_url_system_id",
            name,
            asset_type,
            domain,
            func.coalesce(func.md5(base_url), "NULL"),
            system_id,
            unique=True,
        ),
    )

    @classmethod
    async def upsert_async(
        cls: Type[Asset],
        async_session: AsyncSession,
        *,
        data: Dict[str, Any],
    ) -> Asset:
        """
        Creates a new Asset record if it does not exist, otherwise updates the existing Asset record
        with the attribute values provided in the `data` dict.

        Assets are looked up by the provided attributes that make up their uniqueness criteria:
        - name
        - asset_type
        - domain
        - base_url (if applicable)
        - system_id

        If you provide the ID of an existing asset it will be updated with any provided data
        """
        if "id" not in data and (
            "name" not in data
            or "asset_type" not in data
            or "domain" not in data
            or "system_id" not in data
        ):
            raise ValueError(
                "name, asset_type, domain, and system_id are required fields on assets"
            )

        record_id: str

        if "id" in data:
            result = await async_session.execute(
                select(cls).where(cls.id == data["id"])  # type: ignore[arg-type]
            )
            existing_record = result.scalars().first()
            if existing_record:
                await async_session.execute(
                    update(cls).where(cls.id == existing_record.id).values(data)  # type: ignore[arg-type]
                )
                record_id = existing_record.id
            else:
                raise ValueError(f"Asset with id {data['id']} does not exist")
        else:
            result = await async_session.execute(
                select(cls).where(  # type: ignore[arg-type, call-arg]
                    cls.name == data["name"],
                    cls.asset_type == data["asset_type"],
                    cls.domain == data["domain"],
                    cls.base_url == data.get("base_url"),
                    cls.system_id == data["system_id"],
                )
            )

            existing_record = result.scalars().first()
            if existing_record:
                await async_session.execute(
                    update(cls).where(cls.id == existing_record.id).values(data)  # type: ignore[arg-type]
                )
                record_id = existing_record.id
            else:
                result = await async_session.execute(insert(cls).values(data))  # type: ignore[arg-type]
                record_id = result.inserted_primary_key.id

        result = await async_session.execute(select(cls).where(cls.id == record_id))  # type: ignore[arg-type]
        return result.scalars().first()

    @classmethod
    async def get_by_system_async(
        cls: Type[Asset],
        async_session: AsyncSession,
        system_id: Optional[str] = None,
        system_fides_key: Optional[str] = None,
    ) -> list[Asset]:
        """
        Retrieves all assets associated with a given system,
        using the provided system `id` or `fides_key`, whichever is provided
        """
        if system_id:
            query = select(cls).where(cls.system_id == system_id)  # type: ignore[arg-type]
        else:
            if not system_fides_key:
                raise ValueError(
                    "Either system_id or system_fides_key must be provided"
                )
            query = (
                select(cls)  # type: ignore[arg-type]
                .join(System, System.id == cls.system_id)  # type: ignore[attr-defined]
                .where(System.fides_key == system_fides_key)
            )

        result = await async_session.execute(query)
        return result.scalars().all()



# pylint: disable=R0401, C0302

from __future__ import annotations

from enum import Enum as EnumType
from typing import TYPE_CHECKING

from sqlalchemy import Column, ForeignKey, String
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import relationship
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.cryptography.cryptographic_util import (
    hash_credential_with_salt,
    hash_value_with_salt,
)
from fides.api.cryptography.identity_salt import get_identity_salt
from fides.api.db.base_class import Base  # type: ignore[attr-defined]
from fides.api.db.base_class import JSONTypeOverride
from fides.api.migrations.hash_migration_mixin import HashMigrationMixin
from fides.api.schemas.redis_cache import Identity, LabeledIdentity, MultiValue
from fides.config import CONFIG

if TYPE_CHECKING:
    from fides.api.models.privacy_request.consent import Consent, ConsentRequest
    from fides.api.models.privacy_request.privacy_request import PrivacyRequest


class ProvidedIdentityType(EnumType):
    """Enum for privacy request identity types"""

    email = "email"
    phone_number = "phone_number"
    ga_client_id = "ga_client_id"
    ljt_readerID = "ljt_readerID"
    fides_user_device_id = "fides_user_device_id"
    external_id = "external_id"


class ProvidedIdentity(HashMigrationMixin, Base):  # pylint: disable=R0904
    """
    A table for storing identity fields and values provided at privacy request
    creation time.
    """

    privacy_request_id = Column(
        String,
        ForeignKey("privacyrequest.id", ondelete="CASCADE", onupdate="CASCADE"),
    )
    privacy_request = relationship(
        "PrivacyRequest",
        backref="provided_identities",
    )  # Which privacy request this identity belongs to

    field_name = Column(
        String,
        index=False,
        nullable=False,
    )
    field_label = Column(
        String,
        index=False,
        nullable=True,
    )
    hashed_value = Column(
        String,
        index=True,
        unique=False,
        nullable=True,
    )  # This field is used as a blind index for exact match searches
    encrypted_value = Column(
        MutableDict.as_mutable(
            StringEncryptedType(
                JSONTypeOverride,
                CONFIG.security.app_encryption_key,
                AesGcmEngine,
                "pkcs5",
            )
        ),
        nullable=True,
    )  # Type bytea in the db
    consent = relationship(
        "Consent", back_populates="provided_identity", cascade="delete, delete-orphan"
    )
    consent_request = relationship(
        "ConsentRequest",
        back_populates="provided_identity",
        cascade="delete, delete-orphan",
    )

    @classmethod
    def bcrypt_hash_value(
        cls,
        value: MultiValue,
        encoding: str = "UTF-8",
    ) -> str:
        """
        Temporary function used to hash values to the previously used bcrypt hashes.
        This can be removed once the bcrypt to SHA-256 migration is complete.
        """

        SALT = "$2b$12$UErimNtlsE6qgYf2BrI1Du"
        value_str = str(value)
        hashed_value = hash_credential_with_salt(
            value_str.encode(encoding),
            SALT.encode(encoding),
        )
        return hashed_value

    @classmethod
    def hash_value(
        cls,
        value: MultiValue,
        encoding: str = "UTF-8",
    ) -> str:
        """Utility function to hash the value with a generated salt"""
        SALT = get_identity_salt()
        value_str = str(value)
        hashed_value = hash_value_with_salt(
            value_str.encode(encoding),
            SALT.encode(encoding),
        )
        return hashed_value

    def migrate_hashed_fields(self) -> None:
        if value := self.encrypted_value.get("value"):
            self.hashed_value = self.hash_value(value)
        self.is_hash_migrated = True

    def as_identity_schema(self) -> Identity:
        """Creates an Identity schema from a ProvidedIdentity record in the application DB."""

        identity_dict = {}
        if any(
            [
                not self.field_name,
                not self.encrypted_value,
            ]
        ):
            return Identity()

        value = self.encrypted_value.get("value")  # type:ignore
        if self.field_label:
            value = LabeledIdentity(label=self.field_label, value=value)
        identity_dict[self.field_name] = value
        return Identity(**identity_dict)



from __future__ import annotations

import re
from enum import Enum
from typing import Any, Dict, List, Optional, Set, Type

from fideslang.validation import FidesKey, validate_fides_key
from sqlalchemy import Boolean, Column
from sqlalchemy import Enum as EnumColumn
from sqlalchemy import Float, ForeignKey, String, UniqueConstraint, or_, text
from sqlalchemy.dialects.postgresql import ARRAY, JSONB
from sqlalchemy.ext.mutable import MutableList
from sqlalchemy.orm import RelationshipProperty, Session, relationship
from sqlalchemy.orm.dynamic import AppenderQuery
from sqlalchemy.util import hybridproperty

from fides.api.db.base_class import Base, FidesBase
from fides.api.models import (
    create_historical_data_from_record,
    dry_update_data,
    update_if_modified,
)
from fides.api.models.asset import Asset
from fides.api.models.location_regulation_selections import DeprecatedNoticeRegion
from fides.api.models.sql_models import (  # type: ignore[attr-defined]
    System,
    get_system_data_uses,
)
from fides.api.schemas.language import SupportedLanguage


class PrivacyNoticeFramework(Enum):
    gpp_us_national = "gpp_us_national"
    gpp_us_state = "gpp_us_state"


class UserConsentPreference(Enum):
    opt_in = "opt_in"  # The user wants to opt in to the notice
    opt_out = "opt_out"  # The user wants to opt out of the notice
    acknowledge = "acknowledge"  # The user has acknowledged this notice
    tcf = "tcf"  # Overall preference set for TCF where there are numerous preferences under the single notice


class ConsentMechanism(Enum):
    """
    Enum is not formalized in the DB because it may be subject to frequent change
    """

    opt_in = "opt_in"
    opt_out = "opt_out"
    notice_only = "notice_only"


class EnforcementLevel(Enum):
    """
    Enum is not formalized in the DB because it may be subject to frequent change
    """

    frontend = "frontend"
    system_wide = "system_wide"
    not_applicable = "not_applicable"


class PrivacyNoticeBase:
    """
    This class contains the common fields between PrivacyNoticeTemplate, PrivacyNotice, and PrivacyNoticeHistory.
    These fields are not translated.
    """

    consent_mechanism = Column(EnumColumn(ConsentMechanism), nullable=False)
    data_uses = Column(
        ARRAY(String),
        nullable=False,
        server_default="{}",
        default=dict,
    )  # a list of `fides_key`s of `DataUse` records

    disabled = Column(Boolean, nullable=False, default=False)
    enforcement_level = Column(EnumColumn(EnforcementLevel), nullable=False)
    framework = Column(String)
    gpp_field_mapping = Column(MutableList.as_mutable(JSONB), index=False, unique=False)
    has_gpc_flag = Column(Boolean, nullable=False, default=False)
    internal_description = Column(String)  # Visible to internal users only
    name = Column(String, nullable=False)
    notice_key = Column(String, nullable=False)

    @property
    def is_gpp(self) -> bool:
        return self.framework in (
            PrivacyNoticeFramework.gpp_us_national.value,
            PrivacyNoticeFramework.gpp_us_state.value,
        )

    def applies_to_system(self, system: System) -> bool:
        """Privacy Notice applies to System if a data use matches or the Privacy Notice
        Data Use is a parent of a System Data Use
        """
        for system_data_use in System.get_data_uses([system], include_parents=True):
            for privacy_notice_data_use in self.data_uses or []:
                if system_data_use == privacy_notice_data_use:
                    return True
        return False

    @classmethod
    def generate_notice_key(cls, name: Optional[str]) -> FidesKey:
        """Generate a notice key from a notice name"""
        if not isinstance(name, str):
            raise Exception("Privacy notice keys must be generated from a string.")
        notice_key: str = re.sub(r"\s+", "_", name.lower().strip())
        return FidesKey(validate_fides_key(notice_key))


class PrivacyNoticeTemplate(PrivacyNoticeBase, Base):
    """
    This table contains the out-of-the-box Privacy Notice Templates that are shipped with Fides
    """

    # All out-of-the-box Notice Translations, stored as a list of JSON
    translations = Column(ARRAY(JSONB))

    def dry_update(self, *, data: dict[str, Any]) -> FidesBase:
        """
        A utility method to dry update the PrivacyNoticeTemplate

        This is used to see what an object update would look like, in memory,
        without actually persisting the update to the db
        """
        updated_attributes: Dict = dry_update_data(resource=self, data_updates=data)

        # create a new object with the updated attribute data to keep this
        # ORM object (i.e., `self`) pristine
        return PrivacyNoticeTemplate(**updated_attributes)


class PrivacyNotice(PrivacyNoticeBase, Base):
    """
    A notice set up by a system administrator that an end user (i.e., data subject)
    accepts or rejects to indicate their consent for particular data uses
    """

    id = Column(
        String(255), primary_key=True, index=True, default=FidesBase.generate_uuid
    )

    origin = Column(
        String,
        ForeignKey(PrivacyNoticeTemplate.id_field_path),
    )  # pointer back to the PrivacyNoticeTemplate

    translations: RelationshipProperty[List[NoticeTranslation]] = relationship(
        "NoticeTranslation",
        backref="privacy_notice",
        lazy="selectin",
        order_by="NoticeTranslation.created_at",
    )

    parent_id = Column(
        String,
        ForeignKey("privacynotice.id", ondelete="SET NULL"),
        nullable=True,
    )
    children: "RelationshipProperty[List[PrivacyNotice]]" = relationship(
        "PrivacyNotice",
        back_populates="parent",
        cascade="all",
        passive_deletes=True,
    )
    parent: "RelationshipProperty[Optional[PrivacyNotice]]" = relationship(
        "PrivacyNotice",
        back_populates="children",
        remote_side=[id],
    )

    @hybridproperty
    def default_preference(self) -> UserConsentPreference:
        """Returns the user's default consent preference given the consent
        mechanism of this notice, or "what is granted to the user"

        For example, if a notice has an opt in consent mechanism, this means
        that they should be granted the opportunity to opt in, but by
        default, they *should be opted out*
        """
        if self.consent_mechanism == ConsentMechanism.opt_in:
            return UserConsentPreference.opt_out  # Intentional
        if self.consent_mechanism == ConsentMechanism.opt_out:
            return UserConsentPreference.opt_in  # Intentional
        if self.consent_mechanism == ConsentMechanism.notice_only:
            return UserConsentPreference.acknowledge

        raise Exception("Invalid notice consent mechanism.")

    @property
    def cookies(self) -> List[Asset]:
        """Return relevant assets of type 'cookie' (via the data use)"""
        db = Session.object_session(self)
        or_queries = [
            f"array_to_string(data_uses, ',') ILIKE '{data_use}%'"
            for data_use in self.data_uses
        ]

        query = db.query(Asset).filter(
            Asset.asset_type == "Cookie",
            or_(*[text(query) for query in or_queries]),
        )

        return query.all()

    @property
    def calculated_systems_applicable(self) -> bool:
        """Convenience property to return if any systems overlap with this notice's data uses

        This is used in the Plus API
        """
        db = Session.object_session(self)

        all_system_data_uses: Set[str] = get_system_data_uses(db, include_parents=True)
        return bool(set(self.data_uses).intersection(all_system_data_uses))

    @property
    def configured_regions_for_notice(self) -> List[str]:
        """Convenience property to look up which regions are using these Notices.

        This is used in the Plus API
        """
        db = Session.object_session(self)

        # Raw sql for performance
        experience_regions_cursor_result = db.execute(
            text(
                """
            SELECT array_agg(distinct(privacyexperience.region) order by privacyexperience.region) AS regions
            FROM privacynotice
               JOIN experiencenotices
                 ON experiencenotices.notice_id = :privacy_notice_id
               JOIN privacyexperienceconfig
                 ON privacyexperienceconfig.id = experiencenotices.experience_config_id
               JOIN privacyexperience
                 ON privacyexperience.experience_config_id = privacyexperienceconfig.id
            GROUP BY privacynotice.id
            """
            ),
            {"privacy_notice_id": self.id},
        )
        res = [result for result in experience_regions_cursor_result]
        return res[0]["regions"] if res else []

    @classmethod
    def fetch_and_validate_notices(
        cls,
        db: Session,
        child_notices: List[Dict[str, Any]],
        parent_id: Optional[str] = None,
    ) -> List[PrivacyNotice]:
        """
        Retrieves existing notices from the database and validates their linkage status.

        This function performs two main checks:
        1. Ensures all specified notices exist in the database.
        2. Verifies that notices are not linked to a different parent than the one specified.
        """

        notice_map = {
            child_notice["id"]: child_notice["name"] for child_notice in child_notices
        }
        child_notice_ids = set(notice_map.keys())
        existing_notices: List[PrivacyNotice] = (
            db.query(cls).filter(cls.id.in_(child_notice_ids)).all()
        )

        if missing_ids := child_notice_ids - {notice.id for notice in existing_notices}:
            missing_names = [notice_map[id] for id in missing_ids]
            raise ValueError(
                f"The following notices do not exist: {', '.join(missing_names)}"
            )

        if linked_ids := {
            notice.id
            for notice in existing_notices
            if notice.parent_id is not None and notice.parent_id != parent_id
        }:
            linked_names = [notice_map[id] for id in linked_ids]
            raise ValueError(
                f"The following notices are already linked to another notice: {', '.join(linked_names)}"
            )

        return existing_notices

    @classmethod
    def create(
        cls: Type[PrivacyNotice],
        db: Session,
        *,
        data: dict[str, Any],
        check_name: bool = False,
    ) -> PrivacyNotice:
        """
        Creates a Privacy Notice and then a NoticeTranslation record for each supplied Translation.

        - For each Notice translation, we also create a PrivacyNoticeHistory record which versions details
        from both the Notice and the Translation itself for auditing purposes.
        - Privacy preferences can be saved separately against this PrivacyNoticeHistory record which contains all the info
        about the notice and the translation to which the user supplied a consent preference.
        """
        translations = data.pop("translations", []) or []
        child_notices = data.pop("children", []) or []

        # return an error before the actual creation if the children are not valid
        existing_notices = cls.fetch_and_validate_notices(db, child_notices)

        created: PrivacyNotice = super().create(db=db, data=data, check_name=check_name)

        # automatically links and unlinks at the DB level with this assignment
        created.children = existing_notices
        db.commit()

        data.pop(
            "id", None
        )  # Default Notices have id specified but we don't want to use the same id for the historical record

        for translation_data in translations:
            # Create the Notice Translation
            translation = NoticeTranslation.create(
                db, data={**translation_data, "privacy_notice_id": created.id}
            )
            # Take a snapshot of the Privacy Notice and the Notice Translation combined. Privacy preferences
            # are saved against this record.
            PrivacyNoticeHistory.create(
                db,
                data={**data, **translation_data, "translation_id": translation.id},
                check_name=False,
            )

        return created

    def update(self, db: Session, *, data: dict[str, Any]) -> PrivacyNotice:
        """
        Updates the Privacy Notice, child privacy notices, and its translations.

        - Links or unlinks child privacy notices to match the children in the request
        - Upserts or deletes supplied translations to match translations in the request
        - For each remaining translation, create a historical record if the base notice
        or translation changed.
        """
        request_translations = data.pop("translations", [])
        child_notices = data.pop("children", [])

        existing_notices = PrivacyNotice.fetch_and_validate_notices(
            db, child_notices, self.id
        )
        # automatically links and unlinks at the DB level with this assignment
        self.children = existing_notices

        # Performs a patch update of the base privacy notice
        base_notice_updated: bool = update_if_modified(self, db=db, data=data)

        for translation_data in request_translations:
            existing_translation: Optional[NoticeTranslation] = (
                self.get_translation_by_language(db, translation_data.get("language"))
            )

            if existing_translation:
                translation: NoticeTranslation = existing_translation
                # Performs a patch update of the existing translation
                translation_updated: bool = update_if_modified(  # type: ignore[attr-defined]
                    existing_translation,
                    db=db,
                    data={**translation_data, "privacy_notice_id": self.id},
                )
            else:
                translation_updated = True
                # Creates a new translation
                translation = NoticeTranslation.create(
                    db,
                    data={**translation_data, "privacy_notice_id": self.id},
                )

            if base_notice_updated or translation_updated:
                create_historical_record_for_notice_and_translation(
                    db=db, privacy_notice=self, notice_translation=translation
                )

        # Removes any translations not supplied in the request from the Notice
        delete_notice_translations(
            db, privacy_notice=self, request_translations=request_translations
        )

        return self  # type: ignore[return-value]

    def dry_update(self, *, data: dict[str, Any]) -> PrivacyNotice:
        """
        A utility method to get an updated object without saving it to the db.

        This is used to see what an object update would look like, in memory,
        without actually persisting the update to the db
        """
        updated_attributes: Dict = dry_update_data(resource=self, data_updates=data)

        # This relationship needs to be removed before the dry_update
        # to prevent the dry update from being added to Session.new
        updated_attributes.pop("translations", [])
        updated_attributes.pop("children", [])

        # create a new object with the updated attribute data to keep this
        # ORM object (i.e., `self`) pristine
        return PrivacyNotice(**updated_attributes)

    def get_translation_by_language(
        self, db: Session, language: Optional[SupportedLanguage]
    ) -> Optional[NoticeTranslation]:
        """Lookup a translation on a Privacy Notice by language if it exists"""
        if not language:
            # Shouldn't be possible, but just in case
            return None
        return (
            db.query(NoticeTranslation)
            .filter(
                NoticeTranslation.language == language,
                NoticeTranslation.privacy_notice_id == self.id,
            )
            .first()
        )


class NoticeTranslationBase:
    """Base fields for Notice Translations
    These fields are translated
    """

    description = Column(String)
    language = Column(
        EnumColumn(
            SupportedLanguage,
            native_enum=False,
            values_callable=lambda x: [i.value for i in x],
        ),
    )
    title = Column(String, nullable=False)


class NoticeTranslation(NoticeTranslationBase, Base):
    """Available translations saved for a given Privacy Notice"""

    language = Column(
        EnumColumn(
            SupportedLanguage,
            native_enum=False,
            values_callable=lambda x: [i.value for i in x],
        ),
        nullable=False,
    )  # Overrides NoticeTranslationBase to make language non-nullable

    privacy_notice_id = Column(
        String, ForeignKey(PrivacyNotice.id_field_path), nullable=False
    )

    histories: RelationshipProperty[AppenderQuery] = relationship(
        "PrivacyNoticeHistory",
        backref="notice_translation",
        lazy="dynamic",
        order_by="PrivacyNoticeHistory.created_at",
    )

    __table_args__ = (
        UniqueConstraint("language", "privacy_notice_id", name="notice_translation"),
    )

    @property
    def version(self) -> Optional[float]:
        """Convenience property that returns the latest translation version number for this notice"""
        return (
            self.privacy_notice_history.version if self.privacy_notice_history else None  # type: ignore[return-value]
        )

    @property
    def privacy_notice_history(self) -> Optional[PrivacyNoticeHistory]:
        """Convenience property that returns the privacy notice history for the latest version.

        Note that there are possibly many historical records for the notice translation, this just returns the current
        corresponding historical record.
        """
        # Histories are sorted at the relationship level
        return self.histories[-1] if self.histories.count() else None

    @property
    def privacy_notice_history_id(self) -> Optional[str]:
        """Convenience property that returns the privacy notice history id for the current version.

        Note that there are possibly many historical records for the given notice translation, this just returns the current
        corresponding historical record.
        """
        return self.privacy_notice_history.id if self.privacy_notice_history else None


class DeprecatedPrivacyNoticeHistoryFields:
    """
    Fields we no longer save on the privacy notice, but are now configured on the Experience side.

    However, these fields are retained for auditing purposes on early historical records.
    """

    displayed_in_privacy_center = Column(
        Boolean,
    )
    displayed_in_overlay = Column(
        Boolean,
    )
    displayed_in_api = Column(
        Boolean,
    )
    regions = Column(
        ARRAY(EnumColumn(DeprecatedNoticeRegion, native_enum=False)),
        index=True,
    )


class PrivacyNoticeHistory(
    NoticeTranslationBase, PrivacyNoticeBase, DeprecatedPrivacyNoticeHistoryFields, Base
):
    """
    An "audit table" stores all versions of `PrivacyNotice` + `NoticeTranslations`.

    - When a Privacy Notice and/or its translations are updated, each translation has a PrivacyNoticeHistory record created.
    Privacy preferences are saved against this notice history.
    """

    origin = Column(
        String,
        ForeignKey(PrivacyNoticeTemplate.id_field_path),
    )  # pointer back to the PrivacyNoticeTemplate

    translation_id = Column(
        String,
        ForeignKey(NoticeTranslation.id_field_path, ondelete="SET NULL"),
        index=True,
    )  # pointer back to the NoticeTranslation.  Set to null if the translation is deleted, but
    # we retain this record for consent reporting

    version = Column(Float, nullable=False, default=1.0)


def create_historical_record_for_notice_and_translation(
    db: Session, privacy_notice: PrivacyNotice, notice_translation: NoticeTranslation
) -> None:
    existing_version: float = notice_translation.version or 0.0
    history_data: dict = create_historical_data_from_record(privacy_notice)
    history_data.pop("translations", None)
    history_data.pop("parent_id", None)

    updated_translation_data: dict = create_historical_data_from_record(
        notice_translation
    )
    # Translations have a FK back to privacy_notice_id, but this historical records do not
    updated_translation_data.pop("privacy_notice_id")

    # Creates a historical record of the Notice and the translation combined. Preferences are saved
    # against this resource.
    PrivacyNoticeHistory.create(
        db,
        data={
            **history_data,
            **updated_translation_data,
            "translation_id": notice_translation.id,
            "version": existing_version + 1.0,
        },
        check_name=False,
    )


def delete_notice_translations(
    db: Session,
    privacy_notice: PrivacyNotice,
    request_translations: Dict,
) -> None:
    """Removes any translations that are currently stored on the PrivacyNotice but not in the update request"""
    notice_translations: List[NoticeTranslation] = privacy_notice.translations
    translations_to_remove: Set[SupportedLanguage] = set(  # type: ignore[assignment]
        translation.language for translation in notice_translations
    ).difference(
        set(translation.get("language") for translation in request_translations)
    )

    db.query(NoticeTranslation).filter(
        NoticeTranslation.language.in_(translations_to_remove),
        NoticeTranslation.privacy_notice_id == privacy_notice.id,
    ).delete()
    db.commit()



"""Changes the email template variables syntax from {{variable_name}} to __VARIABLE_NAME__.

Revision ID: eef4477c37d0
Revises: cc37edf20859
Create Date: 2024-09-03 12:47:54.708196

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "eef4477c37d0"
down_revision = "cc37edf20859"
branch_labels = None
depends_on = None


VARIABLES = ["minutes", "days", "denial_reason", "code", "download_link"]
JSON_VARIABLES = ["subject", "body"]


def upgrade():
    for json_variable in JSON_VARIABLES:
        for variable in VARIABLES:
            statement = f"""
            UPDATE messaging_template
            SET content = jsonb_set(
                content,
                '{{{json_variable}}}',
                to_jsonb(REPLACE(content ->> '{json_variable}', '{{{{{variable}}}}}', '__{variable.upper()}__'))
            );
            """
            op.execute(statement)


def downgrade():
    for json_variable in JSON_VARIABLES:
        for variable in VARIABLES:
            statement = f"""
            UPDATE messaging_template
            SET content = jsonb_set(
                content,
                '{{{json_variable}}}',
                to_jsonb(REPLACE(content ->> '{json_variable}', '__{variable.upper()}__', '{{{{{variable}}}}}'))
            );
            """
            op.execute(statement)



from typing import Any, Dict, Optional, Set

from fideslang.models import Dataset, DatasetField, FidesDatasetReference
from fideslang.validation import FidesKey
from loguru import logger
from sqlalchemy import Column, ForeignKey, String
from sqlalchemy.orm import Session, relationship

from fides.api.common_exceptions import ValidationError
from fides.api.db.base_class import Base
from fides.api.graph.config import (
    Collection,
    CollectionAddress,
    Field,
    FieldAddress,
    FieldPath,
    GraphDataset,
    generate_field,
)
from fides.api.graph.data_type import parse_data_type_string
from fides.api.models.connectionconfig import ConnectionConfig, ConnectionType
from fides.api.service.masking.strategy.masking_strategy import MaskingStrategy
from fides.api.util.saas_util import merge_datasets

from fides.api.models.sql_models import (  # type: ignore[attr-defined] # isort: skip
    Dataset as CtlDataset,
)


class DatasetConfig(Base):
    """
    Stores a Fides-annotated dataset that belongs to a ConnectionConfig.

    NOTE: For simplicity, this only stores the dataset's fides_key for indexing,
    and the rest of the dataset is stored as JSON. See fideslang.models.Collection
    for details
    """

    connection_config_id = Column(
        String, ForeignKey(ConnectionConfig.id_field_path), nullable=False
    )
    fides_key = Column(String, index=True, unique=True, nullable=False)
    ctl_dataset_id = Column(
        String, ForeignKey(CtlDataset.id), index=True, nullable=False
    )

    connection_config = relationship(
        "ConnectionConfig",
        back_populates="datasets",
    )

    ctl_dataset = relationship(
        CtlDataset,
        backref="dataset_configs",
    )

    @classmethod
    def create_or_update(cls, db: Session, *, data: Dict[str, Any]) -> "DatasetConfig":  # type: ignore[override]
        """
        Create or update both DatasetConfig and CTL Dataset.
        Updates existing CTL Dataset if found by ID or fides_key, otherwise creates new one.
        """

        def upsert_ctl_dataset(
            dataset_contents: Dict[str, Any],
            existing_ctl_dataset_id: Optional[str] = None,
        ) -> CtlDataset:
            """Create new or update existing CTL dataset."""
            validated_data = Dataset(**dataset_contents).model_dump(mode="json")

            if existing_ctl_dataset_id:
                ctl_dataset = (
                    db.query(CtlDataset)
                    .filter(CtlDataset.id == existing_ctl_dataset_id)
                    .first()
                )
            else:
                ctl_dataset = (
                    db.query(CtlDataset)
                    .filter(CtlDataset.fides_key == dataset_contents.get("fides_key"))
                    .first()
                )

            if ctl_dataset:
                for key, val in validated_data.items():
                    setattr(ctl_dataset, key, val)
            else:
                ctl_dataset = CtlDataset(**validated_data)

            db.add(ctl_dataset)
            db.commit()
            db.refresh(ctl_dataset)
            return ctl_dataset

        # Make a copy of data to avoid modifications
        data_copy = data.copy()
        dataset_contents = data_copy.pop("dataset", None)

        # Check for existing dataset config
        dataset_config = cls.filter(
            db=db,
            conditions=(
                (cls.connection_config_id == data_copy["connection_config_id"])
                & (cls.fides_key == data_copy["fides_key"])
            ),
        ).first()
        existing_ctl_dataset_id = (
            dataset_config.ctl_dataset_id if dataset_config else None
        )

        # Handle CTL dataset if dataset data is provided
        if dataset_contents:
            ctl_dataset = upsert_ctl_dataset(dataset_contents, existing_ctl_dataset_id)
            data_copy["ctl_dataset_id"] = ctl_dataset.id

        # Create or update DatasetConfig
        if dataset_config:
            dataset_config.update(db=db, data=data_copy)
        else:
            dataset_config = cls.create(db=db, data=data_copy)

        return dataset_config

    def get_graph(self) -> GraphDataset:
        """
        Return the saved dataset JSON as a dataset graph for query execution.

        For dataset configs associated to a ConnectionConfig of type saas,
        the corresponding SaaS config is merged in as well
        """
        dataset_graph = convert_dataset_to_graph(
            Dataset.model_validate(self.ctl_dataset), self.connection_config.key  # type: ignore
        )
        if (
            self.connection_config.connection_type == ConnectionType.saas
            and self.connection_config.saas_config is not None
            and self.connection_config.saas_config["fides_key"] == self.fides_key
        ):
            dataset_graph = merge_datasets(
                dataset_graph,
                self.connection_config.get_saas_config().get_graph(self.connection_config.secrets),  # type: ignore
            )
        else:
            logger.debug(
                "Connection config with key {} is not a saas config, skipping merge dataset",
                self.connection_config.key,
            )

        return dataset_graph

    def get_dataset_with_stubbed_collection(self) -> GraphDataset:
        """
        Return a Dataset with a single mock Collection for use in building a graph
        where we only want one node per dataset, instead of one node per collection.  Note that
        the expectation is that there would be no dependencies between nodes on the eventual graph, and the graph
        doesn't require information stored at the collection-level.

        The single Collection will be the resource that gets practically added to the graph, but the intent
        is that this single node represents the overall Dataset, and will execute Dataset-level requests,
        not Collection-level requests.
        """
        dataset_graph = self.get_graph()
        stubbed_collection = Collection(name=dataset_graph.name, fields=[], after=set())

        dataset_graph.collections = [stubbed_collection]
        return dataset_graph

    def get_identities_and_references(self) -> Set[str]:
        """
        Returns all identity and dataset references in the dataset.
        If a field has multiple references only the first reference will be considered.
        """
        result: Set[str] = set()
        dataset: GraphDataset = self.get_graph()
        for collection in dataset.collections:
            # Process the identities in the collection
            result.update(collection.identities().values())
            for _, field_refs in collection.references().items():
                # Take first reference only, we only care that this collection is reachable,
                # how we get there doesn't matter for our current use case
                ref, edge_direction = field_refs[0]
                if edge_direction == "from" and ref.dataset != self.fides_key:
                    result.add(ref.value)
        return result


def to_graph_field(
    field: DatasetField, return_all_elements: Optional[bool] = None
) -> Field:
    """Flattens the dataset field type into its graph representation"""

    # NOTE: on the dataset field, annotations like identity & references are
    # declared on the meta object, whereas in our graph representation these are
    # top-level attributes for convenience

    identity = None
    is_pk = False
    is_array = False
    references = []
    meta_section = field.fides_meta
    sub_fields = []
    length = None
    data_type_name = None
    read_only = None
    custom_request_field = None
    masking_strategy_override = None

    if meta_section:
        identity = meta_section.identity
        if meta_section.primary_key:
            is_pk = meta_section.primary_key
        if meta_section.custom_request_field:
            custom_request_field = meta_section.custom_request_field
        if meta_section.references:
            for reference in meta_section.references:
                # Split the "field" address (e.g. "customers.id") into its component
                # parts: (collection, field). If there are multiple levels of nesting,
                # force these all into the field address.
                #
                # For example, the following dataset field reference:
                # ```
                # references:
                #   - dataset: postgres_example_test_dataset
                #     field: address.id
                # ```
                # becomes: (postgres_example_test_dataset, address, id)
                #
                # Whereas a deeply nested field:
                # references:
                #   - dataset: mongo_example_test_dataset
                #     field: customer_details.extra.meta.created
                # ```
                # becomes: (mongo_example_test_dataset, customer_details, extra.meta.created)
                (ref_collection, *ref_fields) = reference.field.split(".")
                address = FieldAddress(reference.dataset, ref_collection, *ref_fields)
                references.append(
                    (
                        address,
                        (
                            reference.direction.value  # Transforming reference back to a literal first for Pydantic v2
                            if reference.direction
                            else reference.direction
                        ),
                    )
                )
        if meta_section.length is not None:
            # 'if meta_section.length' will not suffice here, we will want to pass through
            # length for any valid integer if it has been set in the config, including 0.
            #
            # Currently 0 is filtered out by validations but better not to filter out 0's
            # here in case we decide to allow it in the future.
            length = meta_section.length

        if meta_section.masking_strategy_override:
            masking_strategy_override = meta_section.masking_strategy_override

        (data_type_name, is_array) = parse_data_type_string(meta_section.data_type)

        if meta_section.return_all_elements:
            # If specified on array field, lifts and passes into sub-fields, for example,
            # arrays of objects
            return_all_elements = True

        if meta_section.read_only:
            read_only = True

    if field.fields:
        sub_fields = [to_graph_field(fld, return_all_elements) for fld in field.fields]
    return generate_field(
        name=field.name,
        data_categories=field.data_categories,
        identity=identity,
        data_type_name=data_type_name,  # type: ignore
        references=references,  # type: ignore
        is_pk=is_pk,
        length=length,
        is_array=is_array,
        sub_fields=sub_fields,
        return_all_elements=return_all_elements,
        read_only=read_only,
        custom_request_field=custom_request_field,
        masking_strategy_override=masking_strategy_override,
    )


def convert_dataset_to_graph(
    dataset: Dataset, connection_key: FidesKey
) -> GraphDataset:
    """
    Converts the given Fides dataset into the concrete graph
    representation needed for query execution
    """

    dataset_name = dataset.fides_key
    after = set()
    if dataset.fides_meta and dataset.fides_meta.after:
        after = set(dataset.fides_meta.after)
    logger.debug("Parsing dataset '{}' into graph representation", dataset_name)
    graph_collections = []
    for collection in dataset.collections:
        collection_skip_processing: bool = bool(
            collection.fides_meta and collection.fides_meta.skip_processing
        )
        if collection_skip_processing:
            logger.debug(
                "Skipping collection {} on dataset {} marked with skip_processing",
                collection.name,
                dataset_name,
            )
            continue
        graph_fields = [to_graph_field(field) for field in collection.fields]
        logger.debug(
            "Parsing dataset {}: parsed collection {} with {} fields",
            dataset_name,
            collection.name,
            len(graph_fields),
        )
        collection_after: Set[CollectionAddress] = set()
        if collection.fides_meta and collection.fides_meta.after:
            collection_after = {
                CollectionAddress(*s.split(".")) for s in collection.fides_meta.after
            }

        collection_erase_after: Set[CollectionAddress] = set()
        if collection.fides_meta and collection.fides_meta.erase_after:
            collection_erase_after = {
                CollectionAddress(*s.split("."))
                for s in collection.fides_meta.erase_after
            }

        masking_override = None
        if collection.fides_meta and collection.fides_meta.masking_strategy_override:
            masking_override = collection.fides_meta.masking_strategy_override

        collection_partitioning = None
        if collection.fides_meta and collection.fides_meta.partitioning:
            collection_partitioning = collection.fides_meta.partitioning

        graph_collection = Collection(
            name=collection.name,
            fields=graph_fields,
            after=collection_after,
            erase_after=collection_erase_after,
            masking_strategy_override=masking_override,
            skip_processing=collection_skip_processing,
            data_categories=(
                set(collection.data_categories) if collection.data_categories else set()
            ),
            partitioning=collection_partitioning,
        )
        graph_collections.append(graph_collection)
    logger.debug(
        "Finished parsing dataset {} with {} collections",
        dataset_name,
        len(graph_collections),
    )

    return GraphDataset(
        name=dataset_name,
        collections=graph_collections,
        connection_key=connection_key,
        after=after,
    )


def validate_dataset_reference(
    db: Session, dataset_reference: FidesDatasetReference
) -> None:
    """
    Validates that the provided FidesDatasetReference refers
    to a `Dataset`, `Collection` and `Field` that actually exist in the DB.
    Raises a `ValidationError` if not.
    """
    dataset_config: Optional[DatasetConfig] = (
        db.query(DatasetConfig)
        .filter(DatasetConfig.fides_key == dataset_reference.dataset)
        .first()
    )
    if dataset_config is None:
        raise ValidationError(
            f"Unknown dataset '{dataset_reference.dataset}' referenced by external reference"
        )

    dataset: GraphDataset = convert_dataset_to_graph(
        Dataset.model_validate(dataset_config.ctl_dataset), dataset_config.fides_key  # type: ignore[arg-type]
    )
    collection_name, *field_name = dataset_reference.field.split(".")
    if not field_name or not collection_name or not field_name[0]:
        raise ValidationError(
            "Dataset reference field specifications must include at least two dot-separated components"
        )
    collection = next(
        (
            collection
            for collection in dataset.collections
            if collection.name == collection_name
        ),
        None,
    )
    if collection is None:
        raise ValidationError(
            f"Unknown collection '{collection_name}' in dataset '{dataset_config.fides_key}' referenced by external reference"
        )
    field = collection.field(FieldPath.parse(*field_name))
    if field is None:
        raise ValidationError(
            f"Unknown field '{dataset_reference.field}' in dataset '{dataset_config.fides_key}' referenced by external reference"
        )


def validate_masking_strategy_override(dataset: Dataset) -> None:
    """
    Validates that field-level masking overrides do not require secret keys.
    When handling a privacy request, we use the `cache_data` function to review the policies and identify which masking strategies need secret keys generated and cached.
    Currently, we are avoiding the additional complexity of scanning datasets for masking overrides.
    """

    def validate_field(dataset_field: DatasetField) -> None:
        if dataset_field.fields:
            for subfield in dataset_field.fields:
                validate_field(subfield)
        else:
            if (
                dataset_field.fides_meta
                and dataset_field.fides_meta.masking_strategy_override
            ):
                strategy: MaskingStrategy = MaskingStrategy.get_strategy(
                    dataset_field.fides_meta.masking_strategy_override.strategy,
                    dataset_field.fides_meta.masking_strategy_override.configuration,  # type: ignore[arg-type]
                )
                if strategy.secrets_required():
                    raise ValidationError(
                        f"Masking strategy '{strategy.name}' with required secrets not allowed as an override."
                    )

    for collection in dataset.collections:
        for field in collection.fields:
            validate_field(field)



import enum

from sqlalchemy import Column
from sqlalchemy import Enum as EnumColumn
from sqlalchemy import String
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.db.base_class import Base
from fides.config import CONFIG


class ProviderEnum(enum.Enum):
    google = "google"
    okta = "okta"
    azure = "azure"
    custom = "custom"


class OpenIDProvider(Base):
    """The DB ORM model for OpenIDProvider."""

    identifier = Column(String, unique=True, index=True)
    name = Column(String)
    provider = Column(EnumColumn(ProviderEnum))
    client_id = Column(
        StringEncryptedType(
            type_in=String(),
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
        nullable=False,
    )
    client_secret = Column(
        StringEncryptedType(
            type_in=String(),
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
        nullable=False,
    )
    domain = Column(String, nullable=True)  # Used for Okta provider
    authorization_url = Column(String, nullable=True)  # Used for Custom provider
    token_url = Column(String, nullable=True)  # Used for Custom provider
    user_info_url = Column(String, nullable=True)  # Used for Custom provider

    @declared_attr
    def __tablename__(self) -> str:
        return "openid_provider"



"""convert field_name to string type

Revision ID: d49a767eb49d
Revises: e4023342ebbb
Create Date: 2024-03-26 04:35:22.358246

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d49a767eb49d"
down_revision = "e4023342ebbb"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column(
        "providedidentity",
        "field_name",
        existing_type=sa.Enum("providedidentitytype"),
        type_=sa.String(),
        existing_nullable=False,
    )
    op.execute("DROP TYPE providedidentitytype")
    op.add_column(
        "providedidentity", sa.Column("field_label", sa.String(), nullable=True)
    )


def downgrade():
    op.drop_column("providedidentity", "field_label")
    op.execute(
        "CREATE TYPE providedidentitytype AS ENUM('email', 'phone_number', 'ga_client_id', 'ljt_readerID', 'fides_user_device_id')"
    )
    op.alter_column(
        "providedidentity",
        "field_name",
        existing_type=sa.String(),
        type_=sa.Enum("providedidentitytype"),
        existing_nullable=False,
    )



# pylint: disable=useless-import-alias

# Because these modules are imported into the __init__.py and used elsewhere they need
# to be explicitly exported in order to prevent implicit reexport errors in mypy.
# This is done by importing "as": `from fides.module import MyClass as MyClass`.

from typing import Any, Dict

from fides.api.models.connectionconfig import ConnectionConfig as ConnectionConfig
from fides.api.models.connectionconfig import ConnectionType as ConnectionType
from fides.api.service.connectors.base_connector import BaseConnector as BaseConnector
from fides.api.service.connectors.bigquery_connector import (
    BigQueryConnector as BigQueryConnector,
)
from fides.api.service.connectors.consent_email_connector import (
    GenericConsentEmailConnector,
)
from fides.api.service.connectors.datahub_connector import (
    DatahubConnector as DatahubConnector,
)
from fides.api.service.connectors.dynamic_erasure_email_connector import (
    DynamicErasureEmailConnector,
)
from fides.api.service.connectors.dynamodb_connector import (
    DynamoDBConnector as DynamoDBConnector,
)
from fides.api.service.connectors.email.attentive_connector import AttentiveConnector
from fides.api.service.connectors.email.sovrn_connector import SovrnConnector
from fides.api.service.connectors.erasure_email_connector import (
    GenericErasureEmailConnector,
)
from fides.api.service.connectors.fides_connector import (
    FidesConnector as FidesConnector,
)
from fides.api.service.connectors.google_cloud_mysql_connector import (
    GoogleCloudSQLMySQLConnector as GoogleCloudSQLMySQLConnector,
)
from fides.api.service.connectors.google_cloud_postgres_connector import (
    GoogleCloudSQLPostgresConnector as GoogleCloudSQLPostgresConnector,
)
from fides.api.service.connectors.http_connector import HTTPSConnector as HTTPSConnector
from fides.api.service.connectors.manual_task_connector import (
    ManualTaskConnector as ManualTaskConnector,
)
from fides.api.service.connectors.manual_webhook_connector import (
    ManualWebhookConnector as ManualWebhookConnector,
)
from fides.api.service.connectors.mariadb_connector import (
    MariaDBConnector as MariaDBConnector,
)
from fides.api.service.connectors.microsoft_sql_server_connector import (
    MicrosoftSQLServerConnector as MicrosoftSQLServerConnector,
)
from fides.api.service.connectors.mongodb_connector import (
    MongoDBConnector as MongoDBConnector,
)
from fides.api.service.connectors.mysql_connector import (
    MySQLConnector as MySQLConnector,
)
from fides.api.service.connectors.okta_connector import OktaConnector as OktaConnector
from fides.api.service.connectors.postgres_connector import (
    PostgreSQLConnector as PostgreSQLConnector,
)
from fides.api.service.connectors.rds_mysql_connector import (
    RDSMySQLConnector as RDSMySQLConnector,
)
from fides.api.service.connectors.rds_postgres_connector import (
    RDSPostgresConnector as RDSPostgresConnector,
)
from fides.api.service.connectors.redshift_connector import (
    RedshiftConnector as RedshiftConnector,
)
from fides.api.service.connectors.s3_connector import S3Connector
from fides.api.service.connectors.saas_connector import SaaSConnector as SaaSConnector
from fides.api.service.connectors.scylla_connector import (
    ScyllaConnector as ScyllaConnector,
)
from fides.api.service.connectors.snowflake_connector import (
    SnowflakeConnector as SnowflakeConnector,
)
from fides.api.service.connectors.timescale_connector import (
    TimescaleConnector as TimescaleConnector,
)
from fides.api.service.connectors.website_connector import WebsiteConnector

supported_connectors: Dict[str, Any] = {
    ConnectionType.attentive_email.value: AttentiveConnector,
    ConnectionType.bigquery.value: BigQueryConnector,
    ConnectionType.datahub.value: DatahubConnector,
    ConnectionType.dynamic_erasure_email.value: DynamicErasureEmailConnector,
    ConnectionType.dynamodb.value: DynamoDBConnector,
    ConnectionType.fides.value: FidesConnector,
    ConnectionType.generic_consent_email.value: GenericConsentEmailConnector,
    ConnectionType.generic_erasure_email.value: GenericErasureEmailConnector,
    ConnectionType.google_cloud_sql_mysql.value: GoogleCloudSQLMySQLConnector,
    ConnectionType.google_cloud_sql_postgres.value: GoogleCloudSQLPostgresConnector,
    ConnectionType.https.value: HTTPSConnector,
    ConnectionType.manual_webhook.value: ManualWebhookConnector,
    ConnectionType.manual_task.value: ManualTaskConnector,
    ConnectionType.mariadb.value: MariaDBConnector,
    ConnectionType.mongodb.value: MongoDBConnector,
    ConnectionType.mssql.value: MicrosoftSQLServerConnector,
    ConnectionType.mysql.value: MySQLConnector,
    ConnectionType.okta.value: OktaConnector,
    ConnectionType.postgres.value: PostgreSQLConnector,
    ConnectionType.rds_mysql.value: RDSMySQLConnector,
    ConnectionType.rds_postgres.value: RDSPostgresConnector,
    ConnectionType.redshift.value: RedshiftConnector,
    ConnectionType.s3.value: S3Connector,
    ConnectionType.saas.value: SaaSConnector,
    ConnectionType.scylla.value: ScyllaConnector,
    ConnectionType.snowflake.value: SnowflakeConnector,
    ConnectionType.sovrn.value: SovrnConnector,
    ConnectionType.timescale.value: TimescaleConnector,
    ConnectionType.website.value: WebsiteConnector,
    ConnectionType.test_website.value: WebsiteConnector,
}


def get_connector(conn_config: ConnectionConfig) -> BaseConnector:
    """Return the Connector class corresponding to the connection_type."""
    try:
        return supported_connectors[conn_config.connection_type.value](conn_config)  # type: ignore
    except KeyError:
        raise NotImplementedError(
            f"Add {conn_config.connection_type} to the 'supported_connectors' mapping."
        )



"""Add privacy notice template

Revision ID: b8248ce6b8a1
Revises: 63b482f5b49b
Create Date: 2023-05-30 14:19:20.629394

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "b8248ce6b8a1"
down_revision = "63b482f5b49b"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "privacynoticetemplate",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("internal_description", sa.String(), nullable=True),
        sa.Column("regions", postgresql.ARRAY(sa.String()), nullable=False),
        sa.Column("consent_mechanism", sa.String(), nullable=False),
        sa.Column("data_uses", postgresql.ARRAY(sa.String()), nullable=False),
        sa.Column("enforcement_level", sa.String(), nullable=False),
        sa.Column("disabled", sa.Boolean(), nullable=False),
        sa.Column("has_gpc_flag", sa.Boolean(), nullable=False),
        sa.Column("displayed_in_privacy_center", sa.Boolean(), nullable=False),
        sa.Column("displayed_in_overlay", sa.Boolean(), nullable=False),
        sa.Column("displayed_in_api", sa.Boolean(), nullable=False),
        sa.Column("notice_key", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_privacynoticetemplate_id"),
        "privacynoticetemplate",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacynoticetemplate_regions"),
        "privacynoticetemplate",
        ["regions"],
        unique=False,
    )
    op.create_foreign_key(
        "privacynotice_origin_fkey",
        "privacynotice",
        "privacynoticetemplate",
        ["origin"],
        ["id"],
    )
    op.create_foreign_key(
        "privacynoticehistory_origin_fkey",
        "privacynoticehistory",
        "privacynoticetemplate",
        ["origin"],
        ["id"],
    )


def downgrade():
    op.drop_constraint(
        "privacynoticehistory_origin_fkey", "privacynoticehistory", type_="foreignkey"
    )
    op.drop_constraint("privacynotice_origin_fkey", "privacynotice", type_="foreignkey")
    op.drop_index(
        op.f("ix_privacynoticetemplate_regions"), table_name="privacynoticetemplate"
    )
    op.drop_index(
        op.f("ix_privacynoticetemplate_id"), table_name="privacynoticetemplate"
    )
    op.drop_table("privacynoticetemplate")



from typing import Any, Dict, List, Optional, Union

import pydash
from loguru import logger
from requests import Response

from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.saas.strategy_configuration import (
    UnwrapPostProcessorConfiguration,
)
from fides.api.service.processors.post_processor_strategy.post_processor_strategy import (
    PostProcessorStrategy,
)


class UnwrapPostProcessorStrategy(PostProcessorStrategy):
    """
    Given a path to a dict/list, returns the dict/list
    E.g.
    data = {
        "exact_matches": {
            "members": [
                {"howdy": 123},
                {"meow": 841}
            ]
        }
    }
    data_path = exact_matches.members
    result = [
                {"howdy": 123},
                {"meow": 841}
            ]

    If given a list, the unwrap will apply to the dicts inside the list.
    """

    name = "unwrap"
    configuration_model = UnwrapPostProcessorConfiguration

    def __init__(self, configuration: UnwrapPostProcessorConfiguration):
        self.data_path = configuration.data_path

    def process(
        self,
        data: Union[List[Dict[str, Any]], Dict[str, Any]],
        identity_data: Optional[Dict[str, Any]] = None,
        privacy_request: Optional[PrivacyRequest] = None,
        response: Optional[Response] = None,
    ) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """
        :param data: A list or dict
        :param identity_data: dict of cached identity data
        :return: unwrapped list, dict, or empty list
        """

        result = []
        if isinstance(data, dict):
            unwrapped = pydash.get(data, self.data_path)
            if unwrapped is None:
                logger.warning(
                    "{} could not be found for the following post processing strategy: {}",
                    self.data_path,
                    self.name,
                )
            else:
                result = unwrapped
        elif isinstance(data, list):
            for item in data:
                unwrapped = pydash.get(item, self.data_path)
                if unwrapped is None:
                    logger.warning(
                        "{} could not be found for the following post processing strategy: {}",
                        self.data_path,
                        self.name,
                    )
                else:
                    result.append(unwrapped)
            # flatten the list to account for the event where the output of unwrapped
            # is a list of lists
            result = pydash.flatten(result)

        return result



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union
from urllib.parse import unquote_to_bytes

from sqlalchemy import Column, ForeignKey, String
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, relationship
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.db.base_class import Base, JSONTypeOverride
from fides.api.util.custom_json_encoder import ENCODED_BYTES_PREFIX
from fides.config import CONFIG

if TYPE_CHECKING:
    from fides.api.models.privacy_request import PrivacyRequest


class MaskingSecret(Base):
    """SQLAlchemy model for storing secret caching information"""

    @declared_attr
    def __tablename__(self) -> str:
        return "masking_secret"

    privacy_request_id = Column(
        String, ForeignKey("privacyrequest.id", ondelete="CASCADE"), nullable=False
    )
    secret = Column(
        StringEncryptedType(
            JSONTypeOverride,
            CONFIG.security.app_encryption_key,
            AesGcmEngine,
            "pkcs5",
        ),
        nullable=False,
    )
    masking_strategy = Column(String, nullable=False)
    secret_type = Column(String, nullable=False)
    privacy_request = relationship("PrivacyRequest", back_populates="masking_secrets")

    def set_secret(self, secret: Union[str, bytes]) -> None:
        """Set the secret value, handling both string and bytes types"""
        if isinstance(secret, str):
            self.secret = secret.encode("utf-8")
        elif isinstance(secret, bytes):
            self.secret = secret
        else:
            raise ValueError("Secret must be either string or bytes")

    def get_secret(self) -> Union[str, bytes]:
        """Retrieve the secret in its original type"""
        secret = self.secret
        if isinstance(secret, str) and secret.startswith(ENCODED_BYTES_PREFIX):
            return unquote_to_bytes(secret)[len(ENCODED_BYTES_PREFIX) :]
        return secret

    @classmethod
    def create(
        cls,
        db: Session,
        *,
        data: dict[str, Any],
        check_name: bool = True,
    ) -> "MaskingSecret":
        """
        Create a new masking secret. Handles both string and bytes secrets automatically, encrypting them for storage.
        """
        return super().create(db=db, data=data, check_name=check_name)



from typing import Any, Dict, List, Optional

from loguru import logger
from sqlalchemy import text
from sqlalchemy.engine import Connection, Engine, create_engine  # type: ignore
from sqlalchemy.orm import Session
from sqlalchemy.sql import Executable  # type: ignore
from sqlalchemy.sql.elements import TextClause

from fides.api.common_exceptions import ConnectionException
from fides.api.graph.execution import ExecutionNode
from fides.api.models.connectionconfig import ConnectionTestStatus
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest, RequestTask
from fides.api.schemas.application_config import SqlDryRunMode
from fides.api.schemas.connection_configuration.connection_secrets_bigquery import (
    BigQuerySchema,
)
from fides.api.service.connectors.query_configs.bigquery_query_config import (
    BigQueryQueryConfig,
)
from fides.api.service.connectors.query_configs.query_config import SQLQueryConfig
from fides.api.service.connectors.sql_connector import SQLConnector
from fides.api.util.collection_util import Row


class BigQueryConnector(SQLConnector):
    """Connector specific to Google BigQuery"""

    secrets_schema = BigQuerySchema

    @property
    def requires_primary_keys(self) -> bool:
        """BigQuery does not have the concept of primary keys so they're not required for erasures."""
        return False

    # Overrides BaseConnector.build_uri
    def build_uri(self) -> str:
        """Build URI of format"""
        config = self.secrets_schema(**self.configuration.secrets or {})
        dataset = f"/{config.dataset}" if config.dataset else ""
        return f"bigquery://{config.keyfile_creds.project_id}{dataset}"  # pylint: disable=no-member

    # Overrides SQLConnector.create_client
    def create_client(self) -> Engine:
        """
        Returns a SQLAlchemy Engine that can be used to interact with Google BigQuery.

        Overrides to pass in credentials_info
        """
        secrets = self.configuration.secrets or {}
        uri = secrets.get("url") or self.build_uri()

        keyfile_creds = secrets.get("keyfile_creds", {})
        credentials_info = dict(keyfile_creds) if keyfile_creds else {}

        return create_engine(
            uri,
            credentials_info=credentials_info,
            hide_parameters=self.hide_parameters,
            echo=not self.hide_parameters,
        )

    # Overrides SQLConnector.query_config
    def query_config(self, node: ExecutionNode) -> BigQueryQueryConfig:
        """Query wrapper corresponding to the input execution_node."""

        db: Session = Session.object_session(self.configuration)
        return BigQueryQueryConfig(
            node, SQLConnector.get_namespace_meta(db, node.address.dataset)
        )

    def partitioned_retrieval(
        self,
        query_config: SQLQueryConfig,
        connection: Connection,
        stmt: TextClause,
    ) -> List[Row]:
        """
        Retrieve data against a partitioned table using the partitioning spec configured for this node to execute
        multiple queries against the partitioned table.

        This is only supported by the BigQueryConnector currently.

        NOTE: when we deprecate `where_clause` partitioning in favor of a more proper partitioning DSL,
        we should be sure to still support the existing `where_clause` partition definition on
        any in-progress DSRs so that they can run through to completion.
        """
        if not isinstance(query_config, BigQueryQueryConfig):
            raise TypeError(
                f"Unexpected query config of type '{type(query_config)}' passed to BigQueryConnector's `partitioned_retrieval`"
            )

        partition_clauses = query_config.get_partition_clauses()
        logger.info(
            f"Executing {len(partition_clauses)} partition queries for node '{query_config.node.address}' in DSR execution"
        )

        if self.should_dry_run(SqlDryRunMode.access):
            for partition_clause in partition_clauses:
                existing_bind_params = stmt.compile().params
                partitioned_stmt = text(
                    f"{stmt} AND ({text(partition_clause)})"
                ).params(existing_bind_params)
                logger.warning(f"SQL DRY RUN - Would execute SQL: {partitioned_stmt}")
            return []

        rows = []
        for partition_clause in partition_clauses:
            logger.debug(
                f"Executing partition query with partition clause '{partition_clause}'"
            )
            existing_bind_params = stmt.compile().params
            partitioned_stmt = text(f"{stmt} AND ({text(partition_clause)})").params(
                existing_bind_params
            )
            results = connection.execute(partitioned_stmt)
            rows.extend(self.cursor_result_to_rows(results))
        return rows

    # Overrides SQLConnector.test_connection
    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """
        Overrides SQLConnector.test_connection with a BigQuery-specific connection test.

        The connection is tested using the native python client for BigQuery, since that is what's used
        by the detection and discovery workflows/codepaths.
        TODO: migrate the rest of this class, used for DSR execution, to also make use of the native bigquery client.
        """
        try:
            bq_schema = BigQuerySchema(**self.configuration.secrets or {})
            client = bq_schema.get_client()
            all_projects = [project for project in client.list_projects()]
            if all_projects:
                return ConnectionTestStatus.succeeded
            logger.error("No Bigquery Projects found with the provided credentials.")
            raise ConnectionException(
                "No Bigquery Projects found with the provided credentials."
            )
        except Exception as e:
            logger.exception(f"Error testing connection to remote BigQuery {str(e)}")
            raise ConnectionException(f"Connection error: {e}")

    def _execute_statements_with_sql_dry_run(
        self,
        statements: List[Executable],
        sql_dry_run_enabled: bool,
        client: Engine,
    ) -> int:
        """
        Execute SQL statements with sql_dry_run support.

        Args:
            statements: List of SQL statements to execute
            sql_dry_run_enabled: Whether sql_dry_run mode is enabled
            client: Database engine

        Returns:
            int: Number of affected rows (0 in sql_dry_run mode)
        """
        if not statements:
            return 0

        if sql_dry_run_enabled:
            for stmt in statements:
                logger.warning(f"SQL DRY RUN - Would execute SQL: {stmt}")
            return 0

        row_count = 0
        with client.connect() as connection:
            for stmt in statements:
                results = connection.execute(stmt)
                logger.debug(f"Affected {results.rowcount} rows")
                row_count += results.rowcount
        return row_count

    def mask_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        rows: List[Row],
        input_data: Optional[Dict[str, List[Any]]] = None,
    ) -> int:
        """Execute a masking request. Returns the number of records updated or deleted"""

        query_config = self.query_config(node)
        update_or_delete_ct = 0
        client = self.client()

        if query_config.uses_delete_masking_strategy():
            delete_stmts = query_config.generate_delete(client, input_data or {})
            logger.debug(f"Generated {len(delete_stmts)} DELETE statements")
            update_or_delete_ct += self._execute_statements_with_sql_dry_run(
                delete_stmts, self.should_dry_run(SqlDryRunMode.erasure), client
            )
        else:
            for row in rows:
                update_or_delete_stmts: List[Executable] = query_config.generate_update(
                    row, policy, privacy_request, client
                )
                logger.debug(
                    f"Generated {len(update_or_delete_stmts)} UPDATE statements"
                )
                update_or_delete_ct += self._execute_statements_with_sql_dry_run(
                    update_or_delete_stmts,
                    self.should_dry_run(SqlDryRunMode.erasure),
                    client,
                )
        return update_or_delete_ct



"""
Contains all of the logic related to the database including connections, setup, teardown, etc.
"""

from os import path
from typing import Literal, Optional, Tuple

from alembic import command, script
from alembic.config import Config
from alembic.runtime import migration
from loguru import logger as log
from sqlalchemy.orm import Session
from sqlalchemy_utils.functions import create_database, database_exists
from sqlalchemy_utils.types.encrypted.encrypted_type import InvalidCiphertextError

from fides.api.db.base import Base  # type: ignore[attr-defined]
from fides.api.db.seed import load_default_resources
from fides.api.db.session import get_db_engine
from fides.api.util.errors import get_full_exception_name

DatabaseHealth = Literal["healthy", "unhealthy", "needs migration"]


def get_alembic_config(database_url: str) -> Config:
    """
    Do lots of magic to make alembic work programmatically from the CLI.
    """

    migrations_dir = path.dirname(path.abspath(__file__))
    directory = path.join(migrations_dir, "../alembic/migrations")
    config = Config(path.join(migrations_dir, "../alembic/alembic.ini"))
    config.set_main_option("script_location", directory.replace("%", "%%"))
    # Avoids invalid interpolation syntax errors if % in string
    config.set_main_option("sqlalchemy.url", database_url.replace("%", "%%"))
    return config


def upgrade_db(alembic_config: Config, revision: str = "head") -> None:
    """Upgrade the database to the specified migration revision."""
    log.info(f"Running database upgrade to revision {revision}")
    command.upgrade(alembic_config, revision)


def init_db(database_url: str) -> None:
    """
    Runs the migrations and creates all of the database objects.
    """
    alembic_config = get_alembic_config(database_url)
    upgrade_db(alembic_config)


def downgrade_db(alembic_config: Config, revision: str = "head") -> None:
    """Downgrade the database to the specified migration revision."""
    log.info(f"Running database downgrade to revision {revision}")
    command.downgrade(alembic_config, revision)


def migrate_db(
    database_url: str,
    revision: str = "head",
    downgrade: bool = False,
) -> None:
    """
    Runs migrations and creates database objects if needed.

    Safe to run on an existing database when upgrading Fides version.
    """
    log.info("Migrating database")
    alembic_config = get_alembic_config(database_url)
    if downgrade:
        downgrade_db(alembic_config, revision)
    else:
        upgrade_db(alembic_config, revision)


def create_db_if_not_exists(database_url: str) -> None:
    """
    Creates a database which does not exist already.
    """
    if not database_exists(database_url):
        create_database(database_url)


def reset_db(database_url: str) -> None:
    """
    Drops all tables/metadata from the database.
    """
    log.info("Resetting database...")
    engine = get_db_engine(database_uri=database_url)
    with engine.connect() as connection:
        log.info("Dropping tables...")
        Base.metadata.drop_all(connection)

        log.info("Dropping Alembic table...")
        migration_context = migration.MigrationContext.configure(connection)
        version = migration_context._version  # pylint: disable=protected-access
        if version.exists(connection):
            version.drop(connection)
    log.info("Reset complete.")


def get_db_health(
    database_url: str, db: Session
) -> Tuple[DatabaseHealth, Optional[str]]:
    """Checks if the db is reachable and up-to-date with Alembic migrations."""
    try:
        alembic_config = get_alembic_config(database_url)
        alembic_script_directory = script.ScriptDirectory.from_config(alembic_config)
        context = migration.MigrationContext.configure(db.connection())
        current_revision = context.get_current_revision()
        if (
            context.get_current_revision()
            != alembic_script_directory.get_current_head()
        ):
            db_health: DatabaseHealth = "needs migration"
        else:
            db_health = "healthy"
        return db_health, current_revision
    except Exception as error:  # pylint: disable=broad-except
        error_type = get_full_exception_name(error)
        log.error("Unable to reach the database: {}: {}", error_type, error)
        return ("unhealthy", None)


def seed_db(session: Session) -> None:
    """Load default resources into the database"""
    load_default_resources(session)


def configure_db(database_url: str, revision: Optional[str] = "head") -> None:
    """Set up the db to be used by the app. Creates db if needed and runs migrations."""
    try:
        create_db_if_not_exists(database_url)
        migrate_db(database_url, revision=revision)  # type: ignore[arg-type]

    except InvalidCiphertextError as cipher_error:
        log.error(
            "Unable to configure database due to a decryption error! Check to ensure your `app_encryption_key` has not changed."
        )
        log.opt(exception=True).error(cipher_error)

    except Exception as error:  # pylint: disable=broad-except
        error_type = get_full_exception_name(error)
        log.error("Unable to configure database: {}: {}", error_type, error)
        log.opt(exception=True).error(error)
        raise


def check_missing_migrations(database_url: str) -> None:
    """
    Tries to autogenerate migrations, returns True if a migration
    was generated.
    """

    engine = get_db_engine(database_uri=database_url)
    connection = engine.connect()

    migration_context = migration.MigrationContext.configure(connection)
    result = command.autogen.compare_metadata(migration_context, Base.metadata)  # type: ignore[attr-defined]

    if result:
        raise SystemExit("Migrations needs to be generated!")
    print("No migrations need to be generated.")



import operator as py_operator
from typing import Any, Union

from loguru import logger
from sqlalchemy.orm import Session

from fides.api.graph.config import FieldPath
from fides.api.task.conditional_dependencies.schemas import (
    Condition,
    ConditionGroup,
    ConditionLeaf,
    GroupOperator,
    Operator,
)

operator_methods = {
    Operator.exists: lambda a, _: a is not None,
    Operator.not_exists: lambda a, _: a is None,
    Operator.eq: py_operator.eq,
    Operator.neq: py_operator.ne,
    Operator.lt: lambda a, b: a < b if a is not None else False,
    Operator.lte: lambda a, b: a <= b if a is not None else False,
    Operator.gt: lambda a, b: a > b if a is not None else False,
    Operator.gte: lambda a, b: a >= b if a is not None else False,
    Operator.list_contains: lambda a, b: b in a if isinstance(a, list) else False,
    Operator.not_in_list: lambda a, b: a not in b if isinstance(b, list) else True,
}


class ConditionEvaluator:
    """Evaluates nested conditions for manual task creation"""

    def __init__(self, db: Session):
        self.db = db

    def evaluate_rule(self, rule: Condition, data: Union[dict, Any]) -> bool:
        """Evaluate a nested condition rule against input data"""
        if isinstance(rule, ConditionLeaf):
            return self._evaluate_leaf_condition(rule, data)
        # ConditionGroup
        return self._evaluate_group_condition(rule, data)

    def _evaluate_leaf_condition(
        self, condition: ConditionLeaf, data: Union[dict, Any]
    ) -> bool:
        """Evaluate a leaf condition against input data"""
        actual_value = self._get_nested_value(data, condition.field_address.split("."))
        # Apply operator and return result
        return self._apply_operator(actual_value, condition.operator, condition.value)

    def _evaluate_group_condition(
        self, group: ConditionGroup, data: Union[dict, Any]
    ) -> bool:
        """Evaluate a group condition against input data"""
        results = [
            self.evaluate_rule(condition, data) for condition in group.conditions
        ]

        logical_operators = {GroupOperator.and_: all, GroupOperator.or_: any}
        operator_func = logical_operators.get(group.logical_operator)

        if operator_func is None:
            logger.warning(f"Unknown logical operator: {group.logical_operator}")
            return False

        return operator_func(results)

    def _get_nested_value(self, data: Union[dict, Any], keys: list[str]) -> Any:
        """Get nested value from data using dot notation

        Supports both simple dictionary access and Fides reference structures:
        - Simple dict: data["user"]["name"]
        - Fides FieldAddress: data.get_field_value(FieldAddress("dataset", "collection", "field_address"))
        - Fides Collection: data.get_field_value(FieldPath("field_address", "subfield"))
        """
        if not keys:
            return data

        current = data

        # Try Fides reference structures first
        if hasattr(current, "get_field_value"):
            try:
                field_path = FieldPath(*keys) if len(keys) > 1 else FieldPath(keys[0])
                return current.get_field_value(field_path)
            except (AttributeError, ValueError):
                pass

        # Fall back to dictionary access
        for key in keys:
            if not isinstance(current, dict):
                current = current.get(key, {}) if hasattr(current, "get") else None
            else:
                current = current.get(key, {})

        return current if current != {} else None

    def _apply_operator(
        self, actual_value: Any, operator: Operator, expected_value: Any
    ) -> bool:
        """Apply operator to actual and expected values"""

        # Get the method for the operator and execute it
        operator_method = operator_methods.get(operator)
        if operator_method is None:
            logger.warning(f"Unknown operator: {operator}")
            return False

        return operator_method(actual_value, expected_value)



from functools import cached_property
from typing import Any, Dict, List, Optional

from botocore.exceptions import ClientError
from loguru import logger
from sqlalchemy.engine import Engine, LegacyCursorResult  # type: ignore
from sqlalchemy.orm import Session

from fides.api.common_exceptions import ConnectionException
from fides.api.graph.execution import ExecutionNode
from fides.api.models.connectionconfig import ConnectionConfig, ConnectionTestStatus
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest, RequestTask
from fides.api.schemas.connection_configuration.connection_secrets_rds_mysql import (
    RDSMySQLSchema,
)
from fides.api.service.connectors.query_configs.mysql_query_config import (
    MySQLQueryConfig,
)
from fides.api.service.connectors.query_configs.query_config import SQLQueryConfig
from fides.api.service.connectors.rds_connector_mixin import RDSConnectorMixin
from fides.api.service.connectors.sql_connector import SQLConnector
from fides.api.util.collection_util import Row


class RDSMySQLConnector(RDSConnectorMixin, SQLConnector):
    """
    Connector specific to RDS MySQL databases
    """

    secrets_schema = RDSMySQLSchema
    namespace_meta: Optional[dict]

    def __init__(self, configuration: ConnectionConfig) -> None:
        super().__init__(configuration)

        self.namespace_meta: Optional[Dict] = None

    def build_uri(self) -> Optional[str]:
        """
        We need to override this method so it is not abstract anymore, and RDSMySQLConnector is instantiable.
        """

    @cached_property
    def typed_secrets(self) -> RDSMySQLSchema:
        return self.secrets_schema(**self.configuration.secrets or {})

    @property
    def url_scheme(self) -> str:
        """
        Returns the URL scheme for the connector's Engine.
        """
        return "mysql+pymysql"

    @property
    def aws_engines(self) -> list[str]:
        """
        Returns the AWS engines supported by the connector.
        """
        return ["mysql", "aurora-mysql"]

    def get_connect_args(self) -> Dict:
        """
        Returns the connection arguments for the Engine.
        """
        connect_args = {
            "ssl": {
                "ca": self.global_bundle_uri,
            }
        }
        return connect_args

    def pre_client_creation_hook(self, node: ExecutionNode) -> None:
        """
        Pre client hook for RDS MySQL Connector
        """
        db: Session = Session.object_session(self.configuration)
        self.namespace_meta = SQLConnector.get_namespace_meta(db, node.address.dataset)

    # Overrides SQLConnector.create_client
    def create_client(self) -> Engine:
        """
        Returns a SQLAlchemy Engine that can be used to interact with a database
        """
        if self.namespace_meta is None:
            raise ConnectionException(
                "Namespace meta is not set. Please call pre_client_creation_hook before creating the client."
            )

        database_instance_name = self.namespace_meta["database_instance_id"]
        db_name = self.namespace_meta["database_id"]
        db_info = self.get_database_instance_connection_info(database_instance_name)
        return self.create_engine(
            db_username=self.typed_secrets.db_username,
            host=db_info["host"],
            port=db_info["port"],
            db_name=db_name,
        )

    def query_config(self, node: ExecutionNode) -> SQLQueryConfig:
        """Query wrapper corresponding to the input execution_node."""
        return MySQLQueryConfig(node)

    def test_connection(self) -> Optional[ConnectionTestStatus]:
        """
        Connects to AWS Account tries to describe RDS instances and clusters.
        """
        logger.info("Starting test connection to {}", self.configuration.key)

        try:
            rds_client = self.rds_client
        except ClientError as e:
            raise ConnectionException(
                f"Error creating RDS client {self.configuration.key}: {e}"
            )

        try:
            logger.info("Describing RDS clusters for {}", self.configuration.key)
            rds_client.describe_db_clusters()
        except ClientError as e:
            raise ConnectionException(
                f"Error describing RDS clusters {self.configuration.key}: {e}"
            )

        try:
            logger.info("Describing RDS instances for {}", self.configuration.key)
            rds_client.describe_db_instances()

        except ClientError as e:
            raise ConnectionException(
                f"Error describing RDS instances {self.configuration.key}: {e}"
            )

        return ConnectionTestStatus.succeeded

    @staticmethod
    def cursor_result_to_rows(results: LegacyCursorResult) -> List[Row]:
        """
        Convert SQLAlchemy results to a list of dictionaries
        """
        return SQLConnector.default_cursor_result_to_rows(results)

    def retrieve_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        input_data: Dict[str, List[Any]],
    ) -> List[Row]:
        """DSR execution not yet supported for RDS MySQL"""
        return []

    def mask_data(
        self,
        node: ExecutionNode,
        policy: Policy,
        privacy_request: PrivacyRequest,
        request_task: RequestTask,
        rows: List[Row],
        input_data: Optional[Dict[str, List[Any]]] = None,
    ) -> int:
        """DSR execution not yet supported for RDS MySQL"""
        return 0



"""unique constraint PreApprovalWebhookReply

Revision ID: c85a641cc92c
Revises: 93af4df40cc0
Create Date: 2024-04-23 19:47:07.836244

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c85a641cc92c"
down_revision = "93af4df40cc0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint(
        "webhook_id_privacy_request_id",
        "preapprovalwebhookreply",
        ["webhook_id", "privacy_request_id"],
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        "webhook_id_privacy_request_id", "preapprovalwebhookreply", type_="unique"
    )
    # ### end Alembic commands ###



from enum import Enum
from typing import Dict, Optional

from fastapi import HTTPException, Security, status
from loguru import logger

from fides.api.api.v1.endpoints import API_PREFIX
from fides.api.db.database import configure_db, migrate_db, reset_db
from fides.api.oauth.utils import verify_oauth_client_prod
from fides.api.util.api_router import APIRouter
from fides.common.api import scope_registry
from fides.config import CONFIG

ADMIN_ROUTER = APIRouter(prefix=API_PREFIX, tags=["Admin"])


class DBActions(str, Enum):
    "The available path parameters for the `/admin/db/{action}` endpoint."

    upgrade = "upgrade"
    reset = "reset"
    downgrade = "downgrade"


@ADMIN_ROUTER.post(
    "/admin/db/{action}",
    tags=["Admin"],
    dependencies=[
        Security(verify_oauth_client_prod, scopes=[scope_registry.DATABASE_RESET])
    ],
    status_code=status.HTTP_200_OK,
)
def db_action(action: DBActions, revision: Optional[str] = "head") -> Dict:
    """
    Initiate one of the enumerated DBActions.

    NOTE: Database downgrades are _not_ guaranteed to succeed, and may put your application
    into an unrecoverable state. They should be invoked with caution. Only downgrade with
    explicit guidance from Ethyca support.
    """

    if action == DBActions.downgrade:
        try:
            migrate_db(database_url=CONFIG.database.sync_database_uri, revision=revision, downgrade=True)  # type: ignore[arg-type]
            action_text = "downgrade"
        except Exception as e:
            logger.exception("Database downgrade failed")
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f"Database downgrade failed: {e}. Check server logs for more details",
            )
    else:
        action_text = "upgrade"

        if action == DBActions.reset:
            if revision != "head":
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail="A revision cannot be specified on 'reset' actions",
                )
            if not CONFIG.dev_mode:
                raise HTTPException(
                    status_code=status.HTTP_501_NOT_IMPLEMENTED,
                    detail="Resetting the application database outside of dev_mode is not supported.",
                )

            reset_db(CONFIG.database.sync_database_uri)
            action_text = "reset"

        try:
            logger.info("Database being configured...")
            configure_db(CONFIG.database.sync_database_uri, revision=revision)
        except Exception as e:
            logger.exception("Database configuration failed: {e}")
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f"Database configuration failed: {e}. Check server logs for more details",
            )

    return {
        "data": {
            "message": f"Fides database action performed successfully: {action_text}"
        }
    }
